from pathlib import Path
from typing import List

from qtpy.QtCore import Qt

from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.core.dialog_box import DialogBox
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.prediction.file_input_widget import (
    PredictionFileInput,
)
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
    PredictionInputMode,
)
from allencell_ml_segmenter.prediction.service import ModelFileService
from allencell_ml_segmenter.core.view import View, MainWindow
from allencell_ml_segmenter.prediction.model_input_widget import (
    ModelInputWidget,
)
from allencell_ml_segmenter.prediction.prediction_folder_progress_tracker import (
    PredictionFolderProgressTracker,
)
from allencell_ml_segmenter.utils.file_utils import FileUtils
from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
    QPushButton,
    QFrame,
    QLabel,
)
from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.core.image_data_extractor import (
    IImageDataExtractor,
    AICSImageDataExtractor,
)


class PredictionView(View, MainWindow):
    """
    Holds the image and model input widgets for prediction.
    """

    def __init__(
        self,
        main_model: MainModel,
        prediction_model: PredictionModel,
        viewer: IViewer,
        img_data_extractor: IImageDataExtractor = AICSImageDataExtractor.global_instance(),
    ):
        super().__init__()
        self._main_model: MainModel = main_model
        self._prediction_model: PredictionModel = prediction_model
        self._viewer: IViewer = viewer
        self._img_data_extractor = img_data_extractor

        self._service: ModelFileService = ModelFileService(
            self._prediction_model
        )

        layout: QVBoxLayout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)
        self.layout().setAlignment(Qt.AlignTop)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)

        self._title: QLabel = QLabel("SEGMENTATION PREDICTION", self)
        self._title.setObjectName("title")
        self.layout().addWidget(self._title, alignment=Qt.AlignHCenter)

        self._file_input_widget: PredictionFileInput = PredictionFileInput(
            self._prediction_model, self._viewer, self._service
        )
        self._file_input_widget.setObjectName("fileInput")

        # Disabled for V1 chrishu 3/30/24 https://github.com/AllenCell/allencell-ml-segmenter/issues/274
        # self._model_input_widget: ModelInputWidget = ModelInputWidget(
        #     self._prediction_model
        # )
        # self._model_input_widget.setObjectName("modelInput")

        # Dummy divs allow for easy alignment
        top_container: QVBoxLayout = QVBoxLayout()
        top_dummy: QFrame = QFrame()
        bottom_container: QVBoxLayout = QVBoxLayout()
        bottom_dummy: QFrame = QFrame()

        top_container.addWidget(self._file_input_widget)
        top_dummy.setLayout(top_container)
        self.layout().addWidget(top_dummy)

        # Disabled for V1 chrishu 3/30/24 https://github.com/AllenCell/allencell-ml-segmenter/issues/274
        # bottom_container.addWidget(self._model_input_widget)

        bottom_dummy.setLayout(bottom_container)
        self.layout().addWidget(bottom_dummy)

        self._run_btn: QPushButton = QPushButton("Run")
        self._run_btn.setObjectName("run")
        self.layout().addWidget(self._run_btn)
        self._run_btn.clicked.connect(self.run_btn_handler)

        self.setStyleSheet(Style.get_stylesheet("prediction_view.qss"))

        self._main_model.subscribe(
            Event.PROCESS_TRAINING_COMPLETE,
            self,
            lambda e: self._main_model.set_current_view(self),
        )

    def run_btn_handler(self):
        # dispatch events to set _prediction_model._input_image_path to a real CSV

        # get image paths from napari if they are selected
        self._prediction_model.dispatch_prediction_get_image_paths_from_napari()
        # Verify prediction is able to start, and write csv if needed
        self._prediction_model.dispatch_prediction_setup()

        total_num_images = self._prediction_model.get_total_num_images()
        if total_num_images:
            progress_tracker: PredictionFolderProgressTracker = (
                PredictionFolderProgressTracker(
                    self._prediction_model.get_output_seg_directory(),
                    total_num_images,
                )
            )
            self.startLongTaskWithProgressBar(progress_tracker)

    def doWork(self):
        self._prediction_model.dispatch_prediction()
        # TODO Need way to set result images to show after prediction complete and refresh viewer.

    def getTypeOfWork(self):
        return "Prediction"

    def showResults(self):
        output_path: Path = self._prediction_model.get_output_seg_directory()

        # Display images if prediction inputs are from Napari Layers
        if (
            self._prediction_model.get_prediction_input_mode()
            == PredictionInputMode.FROM_NAPARI_LAYERS
        ):
            raw_imgs: list[Path] = self._prediction_model.get_selected_paths()
            segmentations: list[Path] = (
                FileUtils.get_all_files_in_dir_ignore_hidden(output_path)
            )
            channel: int = (
                self._prediction_model.get_image_input_channel_index()
            )

            # here, we will pair raw images and segmentations based on the stem component of their paths
            stem_to_data: dict[str, dict[str, Path]] = {
                raw_img.stem: {"raw": raw_img} for raw_img in raw_imgs
            }
            for seg in segmentations:
                # ignore files in the folder that aren't from most recent predictions
                if seg.stem in stem_to_data:
                    stem_to_data[seg.stem]["seg"] = seg

            self._viewer.clear_layers()
            for data in stem_to_data.values():
                self._viewer.add_image(
                    self._img_data_extractor.extract_image_data(
                        data["raw"], channel=channel
                    ).np_data,
                    f"[raw] {data['raw'].name}",
                )
                self._viewer.add_labels(
                    self._img_data_extractor.extract_image_data(
                        data["seg"], seg=1
                    ).np_data,
                    name=f"[seg] {data['seg'].name}",
                )
        # Display popup with saved images path if prediction inputs are from a directory
        else:
            dialog_box = DialogBox(
                f"Predicted images saved to {str(output_path)}. \nWould you like to open this folder?"
            )
            dialog_box.exec()
            if dialog_box.get_selection():
                FileUtils.open_directory_in_window(output_path)

    def focus_changed(self):
        self._viewer.clear_layers()



from allencell_ml_segmenter.sample.sample_model import SampleModel


class SampleProcessService:
    """
    Abstract class for interoperability between fake and real implementation.
    """

    def __init__(self, sample_model: SampleModel):
        self._sample_model = sample_model

    async def run(self):
        """
        Run the service.

        Args:
            process_model (dict): Process model.
            input_data (dict): Input data.

        Returns:
            dict: Output data.
        """
        if self._sample_model.get_process_running():
            return
        elif (
            self._sample_model.get_training_input_files() is None
            or len(self._sample_model.get_training_input_files()) == 0
        ):
            self._sample_model.set_error_message("No Training File Selected")
        else:
            self._sample_model.set_process_running(True)
            for file in self._sample_model.get_training_input_files():
                self._sample_model.append_training_output_files([file])
            self._sample_model.set_process_running(False)



from pathlib import Path
from typing import Optional

from allencell_ml_segmenter.core.image_data_extractor import (
    IImageDataExtractor,
    ImageData,
)
from bioio.bio_image import BioImage

from allencell_ml_segmenter.utils.image_processing import (
    set_all_nonzero_values_to,
)


class AICSImageDataExtractor(IImageDataExtractor):
    """
    Extracts image data using aicsimageio.
    """

    _instance = None

    def extract_image_data(
        self,
        img_path: Path,
        channel: int = 0,
        dims: bool = True,
        np_data: bool = True,
        seg: Optional[int] = None,
    ) -> ImageData:

        aics_img: BioImage = BioImage(img_path)
        if aics_img.dims.T > 1:
            raise RuntimeError("Cannot load timeseries images")

        img_data = None
        if np_data:
            img_data = aics_img.get_image_dask_data("ZYX", C=channel).compute()
            if seg:
                # if this image is a segmentation, replace all values in image with 1 or 2,
                # so it renders correctly as a napari labels layer.
                img_data = set_all_nonzero_values_to(img_data, seg)

        return ImageData(
            aics_img.dims.X if dims else None,
            aics_img.dims.Y if dims else None,
            aics_img.dims.Z if dims else None,
            aics_img.dims.C if dims else None,
            img_data,
            img_path,
        )

    @classmethod
    def global_instance(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



from pathlib import Path
from zipfile import ZipFile

from allencell_ml_segmenter.utils.zip_file import IZipFileManager


class ZipFileManager(IZipFileManager):
    """
    ZipFileManager to manage writing zip files / unzipping zip files
    """

    def __init__(self):
        super().__init__()

    def write_zip_file(self, path: Path, contents: bytes) -> None:
        """
        Write zip file contents :param contents: to :param path:
        """
        # Save file
        with open(path, "wb") as f:
            f.write(contents)

    def unzip_zipped_file_and_delete_zip(self, path_to_zipped: Path) -> None:
        """
        Extract a zipped file to the same directory it is in :param path_to_zipped: 's parent,
        and delete the original zip file at :param path_to_zipped:
        """
        with ZipFile(path_to_zipped, "r") as zipped:
            zipped.extractall(path_to_zipped.parent)
        # delete original zip file
        path_to_zipped.unlink()



from allencell_ml_segmenter.core.directories import Directories
from allencell_ml_segmenter.widgets.input_button_widget import (
    InputButton,
)
from pathlib import Path
from qtpy.QtWidgets import QStackedWidget, QLabel, QSizePolicy
from qtpy.QtGui import QMovie
from qtpy.QtCore import QSize, Qt


class StackedSpinner(QStackedWidget):
    def __init__(self, input_button: InputButton = None):
        super().__init__()

        self.setSizePolicy(
            QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum
        )
        if input_button is not None:
            self.resize(input_button.width(), input_button.height())
        else:
            self.resize(260, 50)

        self.spinner = QLabel(self)
        self.spinner.setAlignment(Qt.AlignCenter)
        gif_path: Path = Directories.get_assets_dir() / "loading.gif"
        self.movie = QMovie(str(gif_path.resolve()))
        self.spinner.setMovie(self.movie)

        self.input_button = input_button
        self.addWidget(self.spinner)
        self.addWidget(self.input_button)
        self.setCurrentWidget(self.input_button)
        self._is_spinning = False

    def start(self):
        self.setCurrentWidget(self.spinner)
        min_dim: int = min(self.spinner.width(), self.spinner.height())
        padding: int = 2  # necessary to make sure the gif isn't cut off
        size: QSize = QSize(min_dim - padding, min_dim - padding)
        # resizing on start in case window size has changed
        self.movie.setScaledSize(size)
        self.movie.start()
        self._is_spinning = True

    def stop(self):
        self.setCurrentWidget(self.input_button)
        self.movie.stop()
        self._is_spinning = False

    def is_spinning(self) -> bool:
        return self._is_spinning



from watchdog.events import FileSystemEvent, FileSystemEventHandler
from typing import Callable


class CacheDirEventHandler(FileSystemEventHandler):
    """
    A CacheDirEventHandler calls :param progress_callback: when a .pt file
    is created in the watched directory with the number of .pt files that have
    been created.
    """

    def __init__(
        self,
        progress_callback: Callable,
    ):
        super().__init__()
        self._progress_callback: Callable = progress_callback
        self._num_files = 0

    # override
    def on_created(self, event: FileSystemEvent) -> None:
        if event.src_path.endswith(".pt"):
            self._num_files += 1
            self._progress_callback(self._num_files)



from pathlib import Path
from typing import List
from unittest.mock import patch, Mock

import pytest
from qtpy.QtWidgets import QFileDialog
from qtpy.QtCore import Qt
from pytestqt.qtbot import QtBot

import allencell_ml_segmenter
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)

from allencell_ml_segmenter.training.image_selection_widget import (
    ImageSelectionWidget,
)
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
)


MOCK_PATH: str = "/path/to/file"


@pytest.fixture
def experiments_model() -> FakeExperimentsModel:
    """
    Fixture that creates an instance of FakeExperimentsModel for testing.
    """
    return FakeExperimentsModel()


@pytest.fixture
def training_model(experiments_model) -> TrainingModel:
    """
    Fixture that creates an instance of TrainingModel for testing.
    """
    return TrainingModel(MainModel(), experiments_model=experiments_model)


@pytest.fixture
def image_selection_widget(
    training_model: TrainingModel, experiments_model: FakeExperimentsModel
) -> ImageSelectionWidget:
    """
    Fixture that creates an instance of ImageSelectionWidget for testing.
    """
    return ImageSelectionWidget(training_model, experiments_model)


# decorator used to stub QFileDialog and avoid nested context managers
@patch.multiple(
    QFileDialog,
    exec_=Mock(return_value=QFileDialog.Accepted),
    selectedFiles=Mock(return_value=[MOCK_PATH]),
)
def test_set_images_directory(
    qtbot: QtBot,
    image_selection_widget: ImageSelectionWidget,
    training_model: TrainingModel,
) -> None:
    """
    Tests that the slot connected to _directory_input_button properly sets the images directory field.
    """
    # ACT
    qtbot.mouseClick(
        image_selection_widget._images_directory_input_button.button,
        Qt.LeftButton,
    )

    # ASSERT
    assert training_model.get_images_directory() == Path(MOCK_PATH)


def test_combos_update_when_num_channels_set(
    qtbot: QtBot,
    image_selection_widget: ImageSelectionWidget,
    training_model: TrainingModel,
) -> None:
    # ASSERT (sanity check)
    assert image_selection_widget._raw_channel_combo_box.count() == 0
    assert image_selection_widget._seg1_channel_combo_box.count() == 0
    assert image_selection_widget._seg2_channel_combo_box.count() == 0

    assert image_selection_widget._raw_channel_combo_box.currentIndex() == -1
    assert image_selection_widget._seg1_channel_combo_box.currentIndex() == -1
    assert image_selection_widget._seg2_channel_combo_box.currentIndex() == -1

    assert not image_selection_widget._raw_channel_combo_box.isEnabled()
    assert not image_selection_widget._seg1_channel_combo_box.isEnabled()
    assert not image_selection_widget._seg2_channel_combo_box.isEnabled()

    assert training_model.get_selected_channel(ImageType.RAW) == None
    assert training_model.get_selected_channel(ImageType.SEG1) == None
    assert training_model.get_selected_channel(ImageType.SEG2) == None

    # ACT (image selection widget should react to this event)
    training_model.set_all_num_channels(
        {
            ImageType.RAW: 1,
            ImageType.SEG1: 2,
            ImageType.SEG2: 3,
        }
    )

    # ASSERT
    assert image_selection_widget._raw_channel_combo_box.count() == 1
    assert image_selection_widget._seg1_channel_combo_box.count() == 2
    assert image_selection_widget._seg2_channel_combo_box.count() == 3

    assert image_selection_widget._raw_channel_combo_box.currentIndex() == 0
    assert image_selection_widget._seg1_channel_combo_box.currentIndex() == 0
    assert image_selection_widget._seg2_channel_combo_box.currentIndex() == 0

    assert image_selection_widget._raw_channel_combo_box.isEnabled()
    assert image_selection_widget._seg1_channel_combo_box.isEnabled()
    assert image_selection_widget._seg2_channel_combo_box.isEnabled()

    assert training_model.get_selected_channel(ImageType.RAW) == 0
    assert training_model.get_selected_channel(ImageType.SEG1) == 0
    assert training_model.get_selected_channel(ImageType.SEG2) == 0


def test_combos_index_controls_model_state(
    qtbot: QtBot,
    image_selection_widget: ImageSelectionWidget,
    training_model: TrainingModel,
) -> None:
    # ARRANGE
    training_model.set_all_num_channels(
        {
            ImageType.RAW: 4,
            ImageType.SEG1: 4,
            ImageType.SEG2: 4,
        }
    )

    # ACT
    image_selection_widget._raw_channel_combo_box.setCurrentIndex(1)
    image_selection_widget._seg1_channel_combo_box.setCurrentIndex(2)
    image_selection_widget._seg2_channel_combo_box.setCurrentIndex(3)

    # ASSERT
    assert training_model.get_selected_channel(ImageType.RAW) == 1
    assert training_model.get_selected_channel(ImageType.SEG1) == 2
    assert training_model.get_selected_channel(ImageType.SEG2) == 3


def test_reset_num_channels_to_none(
    qtbot: QtBot,
    image_selection_widget: ImageSelectionWidget,
    training_model: TrainingModel,
) -> None:
    # ARRANGE
    training_model.set_all_num_channels(
        {
            ImageType.RAW: 4,
            ImageType.SEG1: 4,
            ImageType.SEG2: 4,
        }
    )

    # ASSERT (sanity check)
    assert training_model.get_selected_channel(ImageType.RAW) == 0
    assert training_model.get_selected_channel(ImageType.SEG1) == 0
    assert training_model.get_selected_channel(ImageType.SEG2) == 0

    assert image_selection_widget._raw_channel_combo_box.isEnabled()
    assert image_selection_widget._seg1_channel_combo_box.isEnabled()
    assert image_selection_widget._seg2_channel_combo_box.isEnabled()

    # ACT
    training_model.set_all_num_channels(
        {
            ImageType.RAW: None,
            ImageType.SEG1: None,
            ImageType.SEG2: None,
        }
    )

    # ASSERT
    assert training_model.get_selected_channel(ImageType.RAW) == None
    assert training_model.get_selected_channel(ImageType.SEG1) == None
    assert training_model.get_selected_channel(ImageType.SEG2) == None

    assert not image_selection_widget._raw_channel_combo_box.isEnabled()
    assert not image_selection_widget._seg1_channel_combo_box.isEnabled()
    assert not image_selection_widget._seg2_channel_combo_box.isEnabled()



# CONSTANTS RELATED TO MODEL DOWNLOADS
# Enable Model Downloads on plugin
ENABLE_MODEL_DOWNLOADS = True  # TODO: ENABLE WHEN MODELS UPLOADED
# Endpoint for prod bucket
PROD_BUCKET = (
    "https://production-aics-ml-segmenter-models.s3.us-west-2.amazonaws.com"
)
# Endpoint for stg bucket
STG_BUCKET = (
    "https://staging-aics-ml-segmenter-models.s3.us-west-2.amazonaws.com"
)
# XML namespaces we might expect- currently only aws_s3
# we need this to parse XML with namespaces
# for more info: https://docs.python.org/3/library/xml.etree.elementtree.html#parsing-xml-with-namespaces
XML_NAMESPACES = {"aws_s3": "http://s3.amazonaws.com/doc/2006-03-01/"}



import pytest


@pytest.fixture
def s3_response_listobjectv2_contents_two_models() -> str:
    """
    This function mimics the contents of a response to s3:ListObjectsv2
    if the request is made to a bucket containing two different models and a file.
    example response is from https://docs.aws.amazon.com/AmazonS3/latest/API/API_ListObjectsV2.html
    and does not represent any real data.
    """
    return (
        f'<?xml version="1.0" encoding="UTF-8"?>'
        f'<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
        f"<Name>bucket</Name>"
        f"<Prefix></Prefix>"
        f"<ContinuationToken>randomtoken</ContinuationToken>"
        f"<KeyCount>112</KeyCount>"
        f"<MaxKeys>1000</MaxKeys>"
        f"<IsTruncated>false</IsTruncated>"
        f"  <Contents>"
        f"    <Key>model1.zip</Key>"  # 1st Model file name
        f"    <LastModified>2014-11-21T19:40:05.000Z</LastModified>"
        f'    <ETag>"70ee1738b6b21e2c8a43f3a5ab0eee71"</ETag>'
        f"    <Size>1111</Size>"
        f"    <StorageClass>STANDARD</StorageClass>"
        f"  </Contents>"
        f"  <Contents>"
        f"    <Key>model2.zip</Key>"  # 2nd Model file name
        f"    <LastModified>2014-11-21T19:40:05.000Z</LastModified>"
        f'     <ETag>"70ee1738b6b21e2c8a43f3a5ab0eee71"</ETag>'
        f"    <Size>1111</Size>"
        f"    <StorageClass>STANDARD</StorageClass>"
        f"  </Contents>"
        f"  <Contents>"
        f"    <Key>some_random_file</Key>"  # some random file
        f"    <LastModified>2014-11-21T19:40:05.000Z</LastModified>"
        f'    <ETag>"70ee1738b6b21e2c8a43f3a5ab0eee71"</ETag>'
        f"    <Size>1111</Size>"
        f"    <StorageClass>STANDARD</StorageClass>"
        f"  </Contents>"
        f"</ListBucketResult>"
    )


@pytest.fixture
def s3_response_listobjectv2_contents_duplicate_model() -> str:
    """
    This function mimics the contents of a response to s3:ListObjectsv2
    if the request is made to a bucket containing two models with duplicate names and a file.
    example response is from https://docs.aws.amazon.com/AmazonS3/latest/API/API_ListObjectsV2.html
    and does not represent any real data.
    """
    return (
        f'<?xml version="1.0" encoding="UTF-8"?>'
        f'<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
        f"<Name>bucket</Name>"
        f"<Prefix></Prefix>"
        f"<ContinuationToken>randomtoken</ContinuationToken>"
        f"<KeyCount>112</KeyCount>"
        f"<MaxKeys>1000</MaxKeys>"
        f"<IsTruncated>false</IsTruncated>"
        f"  <Contents>"
        f"    <Key>model1.zip</Key>"  # 1st Model file name
        f"    <LastModified>2014-11-21T19:40:05.000Z</LastModified>"
        f'    <ETag>"70ee1738b6b21e2c8a43f3a5ab0eee71"</ETag>'
        f"    <Size>1111</Size>"
        f"    <StorageClass>STANDARD</StorageClass>"
        f"  </Contents>"
        f"  <Contents>"
        f"    <Key>model1.zip</Key>"  # 2nd Model file name
        f"    <LastModified>2014-11-21T19:40:05.000Z</LastModified>"
        f'     <ETag>"70ee1738b6b21e2c8a43f3a5ab0eee71"</ETag>'
        f"    <Size>1111</Size>"
        f"    <StorageClass>STANDARD</StorageClass>"
        f"  </Contents>"
        f"  <Contents>"
        f"    <Key>some_random_file</Key>"  # some random file
        f"    <LastModified>2014-11-21T19:40:05.000Z</LastModified>"
        f'    <ETag>"70ee1738b6b21e2c8a43f3a5ab0eee71"</ETag>'
        f"    <Size>1111</Size>"
        f"    <StorageClass>STANDARD</StorageClass>"
        f"  </Contents>"
        f"</ListBucketResult>"
    )



from typing import Optional

from qtpy.QtWidgets import QComboBox
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QSizePolicy,
    QLabel,
    QFrame,
    QHBoxLayout,
    QGridLayout,
    QPushButton,
    QProgressBar,
    QRadioButton,
    QDialog,
)
from allencell_ml_segmenter.core.dialog_box import DialogBox
from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.main.viewer import IViewer
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    ImageType,
)
from allencell_ml_segmenter.core.image_data_extractor import ImageData
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint
from allencell_ml_segmenter.curation.stacked_spinner import StackedSpinner
from allencell_ml_segmenter.main.segmenter_layer import ShapesLayer
from allencell_ml_segmenter.core.info_dialog_box import InfoDialogBox
from allencell_ml_segmenter.main.main_model import MIN_DATASET_SIZE


from napari.utils.notifications import show_info, show_warning
from copy import deepcopy
import numpy as np

MERGING_MASK_LAYER_NAME: str = "Merging Mask"
EXCLUDING_MASK_LAYER_NAME: str = "Excluding Mask"


class CurationMainView(QWidget):
    """
    View for Curation UI
    """

    def __init__(self, curation_model: CurationModel, viewer: IViewer) -> None:
        super().__init__()
        self._curation_model: CurationModel = curation_model
        self._viewer: IViewer = viewer

        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 10, 0, 10)
        self.layout().setSpacing(0)
        self.layout().setAlignment(Qt.AlignTop)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        self.setStyleSheet(Style.get_stylesheet("curation_main.qss"))

        self._title: QLabel = QLabel("CURATION UI MAIN", self)
        self._title.setObjectName("title")
        self.layout().addWidget(
            self._title, alignment=Qt.AlignHCenter | Qt.AlignTop
        )

        frame: QFrame = QFrame()
        frame.setLayout(QVBoxLayout())
        self.layout().addWidget(frame)

        input_images_label: QLabel = QLabel("Curation Progress")
        frame.layout().addWidget(input_images_label, alignment=Qt.AlignHCenter)

        progress_bar_layout: QHBoxLayout = QHBoxLayout()
        # Button and progress bar on top row
        # self.back_button: QPushButton = QPushButton("◄ Back")
        # self.back_button.setObjectName("big_blue_btn")
        # progress_bar_layout.addWidget(self.back_button, alignment=Qt.AlignLeft)

        # inner progress bar frame and layout
        inner_progress_frame: QFrame = QFrame()
        inner_progress_frame.setLayout(QHBoxLayout())
        inner_progress_frame.setObjectName("frame")
        self.progress_bar: QProgressBar = QProgressBar()
        inner_progress_frame.layout().addWidget(self.progress_bar)
        self.progress_bar_image_count: QLabel = QLabel("0/0")
        inner_progress_frame.layout().addWidget(
            self.progress_bar_image_count, alignment=Qt.AlignRight
        )
        progress_bar_layout.addWidget(inner_progress_frame)
        self.next_button: QPushButton = QPushButton()
        self.next_button.setObjectName("big_blue_btn")
        self.next_button.clicked.connect(self._on_next)
        progress_bar_layout.addWidget(
            self.next_button, alignment=Qt.AlignRight
        )
        self.layout().addLayout(progress_bar_layout)

        self.save_csv_button: QPushButton = QPushButton(
            "Save Curation Progress"
        )
        self.save_csv_button.clicked.connect(self._on_save_curation_csv)
        self.save_csv_button.setObjectName("save_csv_btn")
        self.layout().addWidget(self.save_csv_button)

        self.file_name: QLabel = QLabel()
        self.layout().addWidget(self.file_name, alignment=Qt.AlignHCenter)

        use_image_frame: QFrame = QFrame()
        use_image_frame.setObjectName("frame")
        use_image_frame.setLayout(QGridLayout())
        use_image_frame.layout().addWidget(
            QLabel("Use this image for training"), 0, 0, 1, 1
        )
        self.yes_radio: QRadioButton = QRadioButton("Yes")
        self.yes_radio.setChecked(True)
        self.yes_radio.clicked.connect(self._on_yes_radio_clicked)

        use_image_frame.layout().addWidget(self.yes_radio, 0, 1, 1, 1)
        self.no_radio: QRadioButton = QRadioButton("No")
        self.no_radio.clicked.connect(self._on_no_radio_clicked)
        use_image_frame.layout().addWidget(self.no_radio, 0, 2, 1, 1)

        self.use_img_help_text = QLabel(
            f"Must select >= {MIN_DATASET_SIZE} images to use"
        )
        self.use_img_help_text.setObjectName("subtext")
        use_image_frame.layout().addWidget(self.use_img_help_text, 2, 0, 1, 1)

        self.use_img_stacked_spinner = StackedSpinner(use_image_frame)
        self.layout().addWidget(
            self.use_img_stacked_spinner, alignment=Qt.AlignHCenter
        )

        optional_text: QLabel = QLabel("OPTIONAL", self)
        optional_text.setObjectName("text_with_vert_padding")
        self.layout().addWidget(optional_text, alignment=Qt.AlignHCenter)

        base_image_layout: QGridLayout = QGridLayout()
        base_combo_label: LabelWithHint = LabelWithHint(
            "Select a base segmentation"
        )
        base_combo_label.set_hint(
            "Areas of the base segmentation that are covered by the merging mask will be overwritten by the other segmentation"
        )
        self.merging_base_combo: QComboBox = QComboBox()
        self.merging_base_combo.addItem("seg1")
        self.merging_base_combo.addItem("seg2")
        self.merging_base_combo.currentIndexChanged.connect(
            lambda idx: self._curation_model.set_base_image(
                self.merging_base_combo.currentText() if idx >= 0 else None
            )
        )

        # these empty QLabels are for spacing... unfortunately cannot apply styling
        # to a QLayout directly, just QWidgets
        base_image_layout.addWidget(QLabel(), 0, 0)
        base_image_layout.addWidget(base_combo_label, 1, 0, 1, 2)
        base_image_layout.addWidget(self.merging_base_combo, 1, 2, 1, 2)
        base_image_layout.addWidget(QLabel(), 2, 0)
        self.layout().addLayout(base_image_layout)
        # Label for Merging mask
        merging_mask_label_and_status: QHBoxLayout = QHBoxLayout()
        merging_mask_label: LabelWithHint = LabelWithHint("Merging mask")
        merging_mask_label.set_hint(
            "Indicates areas of the base segmentation that will be overwritten by the other segmentation"
        )
        self.merging_mask_status: QLabel = QLabel("Create and draw mask")
        merging_mask_label_and_status.addWidget(merging_mask_label)
        merging_mask_label_and_status.addWidget(self.merging_mask_status)
        self.layout().addLayout(merging_mask_label_and_status)
        merging_mask_subtext: QLabel = QLabel(
            "Without merging mask, base image will be used for training."
        )
        merging_mask_subtext.setObjectName("subtext")
        self.layout().addWidget(merging_mask_subtext)

        # buttons for merging mask
        merging_mask_buttons: QHBoxLayout = QHBoxLayout()
        self.merging_create_button: QPushButton = QPushButton("+ Create")
        self.merging_create_button.clicked.connect(self._create_merging_mask)
        self.merging_create_button.setObjectName("small_blue_btn")

        self.merging_delete_button: QPushButton = QPushButton("Delete")
        self.merging_delete_button.clicked.connect(self.delete_merging_mask)

        self.merging_save_button: QPushButton = QPushButton("Save")
        self.merging_save_button.setObjectName("small_blue_btn")
        self.merging_save_button.clicked.connect(self.save_merging_mask)

        merging_mask_buttons.addWidget(self.merging_create_button)
        merging_mask_buttons.addWidget(self.merging_delete_button)
        merging_mask_buttons.addWidget(self.merging_save_button)

        self.layout().addLayout(merging_mask_buttons)

        # Labels for excluding mask
        excluding_mask_labels: QHBoxLayout = QHBoxLayout()
        excluding_mask_labels.setContentsMargins(0, 12, 0, 0)
        excluding_mask_label: LabelWithHint = LabelWithHint("Excluding mask")
        excluding_mask_label.set_hint(
            "Indicates areas of the image set (raw, seg(s)) that will be excluded from training"
        )
        excluding_mask_labels.addWidget(excluding_mask_label)
        self.excluding_mask_status = QLabel("Create and draw mask")
        excluding_mask_labels.addWidget(
            self.excluding_mask_status, alignment=Qt.AlignLeft
        )
        self.layout().addLayout(excluding_mask_labels)

        # buttons for excluding mask
        excluding_mask_buttons: QHBoxLayout = QHBoxLayout()
        self.excluding_create_button: QPushButton = QPushButton("+ Create")
        self.excluding_create_button.setObjectName("small_blue_btn")
        self.excluding_create_button.clicked.connect(
            self._create_excluding_mask
        )
        # propogate button disabled for v1
        # TODO: enable this in v2
        # excluding_propagate_button: QPushButton = QPushButton(
        #    "Propagate in 3D"
        # )
        # excluding_propagate_button.setEnabled(False)

        # TODO: connect the delete button to some functionality
        self.excluding_delete_button: QPushButton = QPushButton("Delete")
        self.excluding_delete_button.clicked.connect(
            self.delete_excluding_mask
        )

        self.excluding_save_button: QPushButton = QPushButton("Save")
        self.excluding_save_button.setObjectName("small_blue_btn")
        self.excluding_save_button.clicked.connect(self.save_excluding_mask)

        excluding_mask_buttons.addWidget(self.excluding_create_button)
        excluding_mask_buttons.addWidget(self.excluding_delete_button)
        excluding_mask_buttons.addWidget(self.excluding_save_button)
        self.layout().addLayout(excluding_mask_buttons)

        self._curation_model.image_loading_finished.connect(
            self._on_first_image_loading_finished
        )

        self._curation_model.saved_to_disk.connect(self._on_saved_to_disk)
        self._set_to_initial_state()

    def _set_to_initial_state(self):
        self.save_csv_button.setEnabled(False)
        self._set_next_button_to_loading()
        self.disable_all_masks()
        self.disable_radio_buttons()
        self.use_img_stacked_spinner.start()

    def _on_image_loading_finished(self) -> None:
        self._enable_next_button()

    def _on_first_image_loading_finished(self) -> None:
        self.use_img_stacked_spinner.stop()
        self.update_save_csv_button_enabled_state()
        self._update_progress_bar()
        self.add_curr_images_to_widget()
        self._enable_next_button()
        self.update_radio_buttons_enabled_state()
        self._curation_model.image_loading_finished.disconnect(
            self._on_first_image_loading_finished
        )
        self._curation_model.image_loading_finished.connect(
            self._on_image_loading_finished
        )

    def _enable_next_button(self) -> None:
        self.next_button.setEnabled(True)
        if self._curation_model.has_next_image():
            self.next_button.setText("Next ►")
        else:
            self.next_button.setText("Finish ►")

    def _set_next_button_to_loading(self) -> None:
        self.next_button.setEnabled(False)
        self.next_button.setText("Loading next...")

    def add_curr_images_to_widget(self) -> None:
        raw_img_data: ImageData = self._curation_model.get_curr_image_data(
            ImageType.RAW
        )
        self._viewer.add_image(
            raw_img_data.np_data, f"[raw] {raw_img_data.path.name}"
        )
        seg1_img_data: ImageData = self._curation_model.get_curr_image_data(
            ImageType.SEG1
        )
        self._viewer.add_labels(
            seg1_img_data.np_data, f"[seg1] {seg1_img_data.path.name}"
        )
        if self._curation_model.has_seg2_data():
            seg2_img_data: ImageData = (
                self._curation_model.get_curr_image_data(ImageType.SEG2)
            )
            self._viewer.add_labels(
                seg2_img_data.np_data, f"[seg2] {seg2_img_data.path.name}"
            )

        self.enable_valid_masks()
        self.file_name.setText(raw_img_data.path.name)
        self.merging_mask_status.setText("Create and draw mask")
        self.excluding_mask_status.setText("Create and draw mask")

    def _on_next(self) -> None:
        """
        Advance to next image set.
        """
        if self.yes_radio.isChecked():
            # prompt and conditionally save unsaved masks
            self._check_unsaved_excluding_mask()
            self._check_unsaved_merging_mask()

        self._viewer.clear_layers()

        if self._curation_model.has_next_image():
            self._set_next_button_to_loading()
            self._curation_model.next_image()
            self.add_curr_images_to_widget()
            # these lines will update UI and model state, must go after
            # a call to next image
            self.yes_radio.click()
            self.update_radio_buttons_enabled_state()
            self.update_save_csv_button_enabled_state()
            self.merging_base_combo.setCurrentIndex(0)
        else:
            self._on_save_curation_csv()
            self.disable_all_masks()
            self.disable_radio_buttons()
            self.file_name.setText("None")
            self.next_button.setEnabled(False)
            self.next_button.setText("No more images")
            self._curation_model.stop_loading_images()
            InfoDialogBox(
                "You have reached the end of the dataset, and your curation CSV has been saved.\nPlease switch to the Training tab to start training a model."
            ).exec()

        self._update_progress_bar()

    def _should_prompt_to_save_mask(
        self,
        mask_layer: Optional[ShapesLayer],
        saved_mask: Optional[np.ndarray],
    ) -> bool:
        if mask_layer is not None:
            return (
                len(mask_layer.data) > 0
                if saved_mask is None
                else not np.array_equal(saved_mask, mask_layer.data)
            )
        return False

    def _check_unsaved_mask(
        self,
        curr_mask_layer: Optional[ShapesLayer],
        saved_mask: Optional[np.ndarray],
        mask_type: str,
    ) -> Optional[np.ndarray]:
        if self._should_prompt_to_save_mask(curr_mask_layer, saved_mask):
            save_changes_prompt = DialogBox(
                f"The current {mask_type} mask layer has unsaved changes. Would you like to save these changes?"
            )
            selection: QDialog.DialogCode = save_changes_prompt.exec()
            if selection == QDialog.DialogCode.Accepted:
                return deepcopy(curr_mask_layer.data)
        return saved_mask

    def _check_unsaved_excluding_mask(self) -> None:
        curr_excl_mask_layer: Optional[ShapesLayer] = self._viewer.get_shapes(
            EXCLUDING_MASK_LAYER_NAME
        )
        saved_excl_mask: Optional[np.ndarray] = (
            self._curation_model.get_excluding_mask()
        )
        self._curation_model.set_excluding_mask(
            self._check_unsaved_mask(
                curr_excl_mask_layer, saved_excl_mask, "excluding"
            )
        )

    def _check_unsaved_merging_mask(self) -> None:
        curr_merg_mask_layer: Optional[ShapesLayer] = self._viewer.get_shapes(
            MERGING_MASK_LAYER_NAME
        )
        saved_merg_mask: Optional[np.ndarray] = (
            self._curation_model.get_merging_mask()
        )
        self._curation_model.set_merging_mask(
            self._check_unsaved_mask(
                curr_merg_mask_layer, saved_merg_mask, "merging"
            )
        )

    def _on_save_curation_csv(self) -> None:
        self._curation_model.save_curr_curation_record_to_disk()
        self.save_csv_button.setEnabled(False)

    def _on_saved_to_disk(self, save_successful: bool) -> None:
        self.save_csv_button.setEnabled(True)
        if save_successful:
            show_info("Current progress saved to CSV")
        else:
            show_warning("Failed to save current progress to CSV")

    def disable_merging_mask_buttons(self):
        """
        Disable the buttons for merging mask in the UI
        """
        self.merging_save_button.setEnabled(False)
        self.merging_create_button.setEnabled(False)
        self.merging_delete_button.setEnabled(False)
        self.merging_base_combo.setEnabled(False)

    def enable_merging_mask_buttons(self):
        """
        Enable the buttons for merging mask in the UI
        """
        mask_exists: bool = self._viewer.contains_layer(
            MERGING_MASK_LAYER_NAME
        )
        self.merging_save_button.setEnabled(mask_exists)
        self.merging_create_button.setEnabled(True)
        self.merging_delete_button.setEnabled(mask_exists)
        self.merging_base_combo.setEnabled(True)

    def disable_excluding_mask_buttons(self):
        """
        disable the buttons for excluding mask in the UI
        """
        self.excluding_save_button.setEnabled(False)
        self.excluding_create_button.setEnabled(False)
        self.excluding_delete_button.setEnabled(False)

    def enable_excluding_mask_buttons(self):
        """
        Enable the buttons for excluding mask in the UI
        """
        mask_exists: bool = self._viewer.contains_layer(
            EXCLUDING_MASK_LAYER_NAME
        )
        self.excluding_save_button.setEnabled(mask_exists)
        self.excluding_create_button.setEnabled(True)
        self.excluding_delete_button.setEnabled(mask_exists)

    def disable_radio_buttons(self):
        self.yes_radio.setEnabled(False)
        self.no_radio.setEnabled(False)

    def update_radio_buttons_enabled_state(self):
        if (
            self._curation_model.get_max_num_images_to_use()
            <= MIN_DATASET_SIZE
        ):
            self.yes_radio.setEnabled(False)
            self.no_radio.setEnabled(False)
            self.use_img_help_text.setText("Must use all remaining images")
        else:
            self.yes_radio.setEnabled(True)
            self.no_radio.setEnabled(True)

    def update_save_csv_button_enabled_state(self):
        self.save_csv_button.setEnabled(
            self._curation_model.get_num_images_selected_to_use()
            >= MIN_DATASET_SIZE
        )

    def _update_progress_bar(self) -> None:
        """
        update progress bar based on state of image loader
        """
        curr_val: int = self._curation_model.get_curr_image_index() + 1
        num_images: int = self._curation_model.get_num_images()
        self.progress_bar.setMaximum(num_images)
        self.progress_bar.setValue(curr_val)
        # set progress bar hint
        self.progress_bar_image_count.setText(f"{curr_val}/{num_images}")

    def _discard_layer_prompt(self, layer: str) -> None:
        discard_layer_prompt = DialogBox(
            f"There is already a '{layer}' layer in the viewer. Would you like to discard this layer?"
        )
        return discard_layer_prompt.exec() == QDialog.DialogCode.Accepted

    def _replace_saved_mask_prompt(self, merging_or_excluding: str):
        replace_prompt = DialogBox(
            f"There is already a {merging_or_excluding} mask layer saved. Would you like to overwrite?"
        )
        return replace_prompt.exec() == QDialog.DialogCode.Accepted

    def _create_merging_mask(self) -> None:
        if self._viewer.contains_layer(MERGING_MASK_LAYER_NAME):
            if not self._discard_layer_prompt(MERGING_MASK_LAYER_NAME):
                return
            self._viewer.remove_layer(MERGING_MASK_LAYER_NAME)

        self._viewer.add_shapes(
            MERGING_MASK_LAYER_NAME, "royalblue", "add_polygon"
        )
        self.merging_save_button.setEnabled(True)
        self.merging_delete_button.setEnabled(True)
        self.merging_mask_status.setText("Draw mask")

    def save_merging_mask(self) -> None:
        if self._curation_model.get_base_image() is None:
            show_info("Please select a base image to merge with")
            return

        merging_mask: Optional[ShapesLayer] = self._viewer.get_shapes(
            MERGING_MASK_LAYER_NAME
        )
        if merging_mask is None:
            show_info("Please create a merging mask layer")
            return

        if self._curation_model.get_merging_mask() is not None:
            if not self._replace_saved_mask_prompt("merging"):
                return

        # deepcopy so that if a user adds more shapes to existing layer, they don't show up in model
        # could change this behavior based on UX input
        self._curation_model.set_merging_mask(deepcopy(merging_mask.data))
        self.merging_mask_status.setText("Merging mask saved")

    def delete_merging_mask(self) -> None:
        merging_mask: Optional[ShapesLayer] = self._viewer.get_shapes(
            MERGING_MASK_LAYER_NAME
        )
        if merging_mask is not None:
            self._viewer.remove_layer(MERGING_MASK_LAYER_NAME)
        self._curation_model.set_merging_mask(None)
        self.merging_save_button.setEnabled(False)
        self.merging_delete_button.setEnabled(False)
        self.merging_mask_status.setText("Merging mask deleted")

    def _create_excluding_mask(self) -> None:
        excluding_mask: Optional[ShapesLayer] = self._viewer.get_shapes(
            EXCLUDING_MASK_LAYER_NAME
        )
        if excluding_mask is not None:
            if not self._discard_layer_prompt(EXCLUDING_MASK_LAYER_NAME):
                return
            self._viewer.remove_layer(EXCLUDING_MASK_LAYER_NAME)

        self._viewer.add_shapes(
            EXCLUDING_MASK_LAYER_NAME, "coral", "add_polygon"
        )
        self.excluding_save_button.setEnabled(True)
        self.excluding_delete_button.setEnabled(True)
        self.excluding_mask_status.setText("Draw mask")

    def save_excluding_mask(self) -> None:
        excluding_mask: Optional[ShapesLayer] = self._viewer.get_shapes(
            EXCLUDING_MASK_LAYER_NAME
        )
        if excluding_mask is None:
            show_info("Please create an excluding mask layer")
            return

        if self._curation_model.get_excluding_mask() is not None:
            if not self._replace_saved_mask_prompt("excluding"):
                return

        self._curation_model.set_excluding_mask(deepcopy(excluding_mask.data))
        self.excluding_mask_status.setText("Excluding mask saved")

    def delete_excluding_mask(self) -> None:
        excluding_mask: Optional[ShapesLayer] = self._viewer.get_shapes(
            EXCLUDING_MASK_LAYER_NAME
        )
        if excluding_mask is not None:
            self._viewer.remove_layer(EXCLUDING_MASK_LAYER_NAME)
        self._curation_model.set_excluding_mask(None)
        self.excluding_save_button.setEnabled(False)
        self.excluding_delete_button.setEnabled(False)
        self.excluding_mask_status.setText("Excluding mask deleted")

    def disable_all_masks(self) -> None:
        self.disable_merging_mask_buttons()
        self.disable_excluding_mask_buttons()

    def enable_valid_masks(self) -> None:
        if self._curation_model.has_seg2_data():
            self.enable_merging_mask_buttons()
        else:
            self.disable_merging_mask_buttons()
        self.enable_excluding_mask_buttons()

    def _on_no_radio_clicked(self) -> None:
        self.disable_all_masks()
        self._curation_model.set_use_image(False)
        self.update_save_csv_button_enabled_state()

    def _on_yes_radio_clicked(self) -> None:
        self.enable_valid_masks()
        self._curation_model.set_use_image(True)
        self.update_save_csv_button_enabled_state()



from pathlib import Path
from typing import List

import pytest

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.prediction.model import PredictionModel
from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber


@pytest.fixture
def prediction_model() -> PredictionModel:
    """
    Fixture that creates an instance of PredictionModel for testing.
    """
    return PredictionModel()


def test_input_image_paths(prediction_model: PredictionModel) -> None:
    """
    Tests that the input image paths are set and retrieved properly.
    """
    # ARRANGE
    dummy_paths: List[Path] = [
        Path("example path " + str(i)) for i in range(10)
    ]

    # ACT
    prediction_model.set_input_image_path(dummy_paths)

    # ASSERT
    assert prediction_model.get_input_image_path() == dummy_paths


def test_image_input_channel_index(prediction_model: PredictionModel) -> None:
    """
    Tests that the channel index is set and retrieved properly.
    """
    for i in range(10):
        # ACT
        prediction_model.set_image_input_channel_index(i)

        # ASSERT
        assert prediction_model.get_image_input_channel_index() == i


def test_output_directory(prediction_model: PredictionModel) -> None:
    """
    Tests that the output directory is set and retrieved properly.
    """
    # ARRANGE
    dummy_path: Path = Path("example path")

    # ACT
    prediction_model.set_output_directory(dummy_path)

    # ASSERT
    assert prediction_model.get_output_directory() == dummy_path


def test_model_path(prediction_model: PredictionModel) -> None:
    """
    Tests that the model path is set and retrieved properly, and that the correct event is fired off.
    """
    # ARRANGE
    event_under_test: Event = Event.ACTION_PREDICTION_MODEL_FILE
    subscriber: FakeSubscriber = FakeSubscriber()
    prediction_model.subscribe(event_under_test, subscriber, subscriber.handle)
    dummy_path: Path = Path("example path")

    # ACT
    prediction_model.set_model_path(dummy_path)

    # ASSERT
    assert prediction_model.get_model_path() == dummy_path
    assert subscriber.was_handled(event_under_test)


def test_preprocessing_method(prediction_model: PredictionModel) -> None:
    """
    Tests that the preprocessing method is set and retrieved properly, and that the correct event is fired off.
    """
    # ARRANGE
    event_under_test: Event = Event.ACTION_PREDICTION_PREPROCESSING_METHOD
    subscriber: FakeSubscriber = FakeSubscriber()
    prediction_model.subscribe(event_under_test, subscriber, subscriber.handle)

    # ACT
    prediction_model.set_preprocessing_method("example method")

    # ASSERT
    assert prediction_model.get_preprocessing_method() == "example method"
    assert subscriber.was_handled(event_under_test)


def test_postprocessing_method(prediction_model: PredictionModel) -> None:
    """
    Tests that the postprocessing method is set and retrieved properly, and that the correct event is fired off.
    """
    # ARRANGE
    event_under_test: Event = Event.ACTION_PREDICTION_POSTPROCESSING_METHOD
    subscriber: FakeSubscriber = FakeSubscriber()
    prediction_model.subscribe(event_under_test, subscriber, subscriber.handle)

    # ACT
    prediction_model.set_postprocessing_method("example method")

    # ASSERT
    assert prediction_model.get_postprocessing_method() == "example method"
    assert subscriber.was_handled(event_under_test)


def test_postprocessing_simple_threshold(
    prediction_model: PredictionModel,
) -> None:
    """
    Tests that the postprocessing simple threshold is set and retrieved properly, and that the correct event is fired off.
    """
    # ARRANGE
    event_under_test: Event = (
        Event.ACTION_PREDICTION_POSTPROCESSING_SIMPLE_THRESHOLD
    )
    subscriber: FakeSubscriber = FakeSubscriber()
    prediction_model.subscribe(event_under_test, subscriber, subscriber.handle)

    # ACT
    prediction_model.set_postprocessing_simple_threshold(0.01)

    # ASSERT
    assert prediction_model.get_postprocessing_simple_threshold() == 0.01
    assert subscriber.was_handled(event_under_test)


def test_postprocessing_auto_threshold(
    prediction_model: PredictionModel,
) -> None:
    """
    Tests that the postprocessing auto threshold is set and retrieved properly, and that the correct event is fired off.
    """
    # ARRANGE
    event_under_test: Event = (
        Event.ACTION_PREDICTION_POSTPROCESSING_AUTO_THRESHOLD
    )
    subscriber: FakeSubscriber = FakeSubscriber()
    prediction_model.subscribe(event_under_test, subscriber, subscriber.handle)

    # ACT
    prediction_model.set_postprocessing_auto_threshold("example threshold")
    assert subscriber.was_handled(event_under_test)

    # ASSERT
    assert (
        prediction_model.get_postprocessing_auto_threshold()
        == "example threshold"
    )



from abc import abstractmethod, ABC
from qtpy.QtWidgets import QWidget, QProgressDialog
from qtpy.QtCore import Qt, QThread

from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.progress_tracker import ProgressTracker


class ViewMeta(type(QWidget), type(Subscriber)):
    pass


class LongTaskThread(QThread):
    def __init__(self, do_work: callable, parent=None):
        super().__init__(parent)
        self._do_work = do_work

    # override
    def run(self):
        print("running")
        self._do_work()


class View(QWidget, Subscriber, metaclass=ViewMeta):
    """
    Base class for all Views to inherit from
    """

    _template = None

    def __init__(self):
        QWidget.__init__(self)

    def startLongTaskWithProgressBar(
        self, progress_tracker: ProgressTracker
    ) -> None:
        self.longTaskThread = LongTaskThread(do_work=self.doWork)

        self.progressDialog = QProgressDialog(
            f"{self.getTypeOfWork()} in Progress",
            "Cancel",
            progress_tracker.get_progress_minimum(),
            progress_tracker.get_progress_maximum(),
            self,
        )
        self.progressDialog.setValue(progress_tracker.get_progress())
        self.progressDialog.setWindowTitle(f"{self.getTypeOfWork()} Progress")
        self.progressDialog.setWindowModality(Qt.ApplicationModal)
        self.progressDialog.canceled.connect(self.longTaskThread.terminate)
        # stop the watchdog thread for file watching inside of the progress tracker
        self.progressDialog.canceled.connect(progress_tracker.stop_tracker)

        self.progressDialog.show()

        self.longTaskThread.finished.connect(self.progressDialog.reset)
        self.longTaskThread.finished.connect(self.longTaskThread.deleteLater)
        self.longTaskThread.finished.connect(self.progressDialog.close)
        self.longTaskThread.finished.connect(self.showResults)

        # connect signals from progress tracker to modify the qprogressdialog
        progress_tracker.signals.progress_changed.connect(self.updateProgress)
        progress_tracker.signals.label_text_changed.connect(
            self.updateLabelText
        )
        progress_tracker.signals.progress_max_changed.connect(
            self.setProgressMax
        )
        # if the longTaskThread or the progressThread finishes, we no longer
        # need to update progress, so we should stop the progress tracker
        self.longTaskThread.finished.connect(progress_tracker.stop_tracker)

        progress_tracker.start_tracker()
        self.longTaskThread.start()

    @abstractmethod
    def showResults(self):
        pass

    def updateProgress(self, value: int) -> None:
        self.progressDialog.setValue(value)

    def updateLabelText(self, value: str) -> None:
        self.progressDialog.setLabelText(value)

    def setProgressMax(self, maximum: int) -> None:
        self.progressDialog.setMaximum(maximum)

    @abstractmethod
    def doWork(self):
        pass

    @abstractmethod
    def getTypeOfWork(self):
        pass


class MainWindow(ABC):
    # this is an ABC that defines a main window in the app, currently this is TrainingView, PredictionView,
    # and CurationMainWidget
    def __init__(self):
        super().__init__()

    @abstractmethod
    def focus_changed(self) -> None:
        pass



from unittest.mock import patch, Mock

import pytest
from qtpy.QtWidgets import QFileDialog
from qtpy.QtCore import Qt
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter.prediction.file_input_widget import (
    PredictionFileInput,
)
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
    PredictionInputMode,
)

MOCK_PATH: str = "/path/to/file"


@pytest.fixture
def prediction_model(qtbot: QtBot) -> PredictionModel:
    return PredictionModel()


@pytest.fixture
def file_input_widget(
    qtbot: QtBot, prediction_model: PredictionModel
) -> PredictionFileInput:
    """
    Fixture that creates an instance of ModelInputWidget for testing.
    """
    return PredictionFileInput(
        prediction_model, viewer=FakeViewer(), service=None
    )


def test_top_radio_button_slot(
    qtbot: QtBot,
    file_input_widget: PredictionFileInput,
    prediction_model: PredictionModel,
) -> None:
    """
    Test the _top_radio_button_slot method of PredictionFileInput.
    """
    # ARRANGE - explicitly disable file_input_widget._image_list and enable file_input_widget._browse_dir_edit
    file_input_widget._image_list.setEnabled(False)
    file_input_widget._browse_dir_edit.setEnabled(True)

    # ACT
    with qtbot.waitSignal(file_input_widget._radio_on_screen.toggled):
        file_input_widget._radio_on_screen.click()

    # ASSERT - states should have flipped
    assert file_input_widget._image_list.isEnabled()
    assert not file_input_widget._browse_dir_edit.isEnabled()
    assert (
        prediction_model.get_prediction_input_mode()
        == PredictionInputMode.FROM_NAPARI_LAYERS
    )


def test_bottom_radio_button_slot(
    qtbot: QtBot,
    file_input_widget: PredictionFileInput,
    prediction_model: PredictionModel,
) -> None:
    """
    Test the _bottom_radio_button_slot method of PredictionFileInput.
    """
    # ARRANGE - explicitly enable file_input_widget._image_list and disable file_input_widget._browse_dir_edit
    file_input_widget._image_list.setEnabled(True)
    file_input_widget._browse_dir_edit.setEnabled(False)

    # ACT
    with qtbot.waitSignal(file_input_widget._radio_directory.toggled):
        file_input_widget._radio_directory.click()

    # ASSERT - states should have flipped
    assert not file_input_widget._image_list.isEnabled()
    assert file_input_widget._browse_dir_edit.isEnabled()
    assert (
        prediction_model.get_prediction_input_mode()
        == PredictionInputMode.FROM_PATH
    )


# decorator used to stub QFileDialog and avoid nested context managers
@patch.multiple(
    QFileDialog,
    exec_=Mock(return_value=QFileDialog.Accepted),
    selectedFiles=Mock(return_value=[MOCK_PATH]),
    getExistingDirectory=Mock(return_value=MOCK_PATH),
)
def test_preprocessing_method(
    qtbot: QtBot, file_input_widget: PredictionFileInput
) -> None:
    """
    Test that the input buttons in file_input_widget do not affect the state related
    to the model_input_widget. This test was introduced because any input button instance
    used to manipulate the model_path state in the prediction model.
    """

    # ACT 1
    qtbot.mouseClick(file_input_widget._browse_dir_edit.button, Qt.LeftButton)

    # ASSERT 1
    assert file_input_widget._model.get_preprocessing_method() is None

    # ACT 2
    qtbot.mouseClick(
        file_input_widget._browse_output_edit.button, Qt.LeftButton
    )

    # ASSERT 2
    assert file_input_widget._model.get_preprocessing_method() is None


def test_populate_input_channel_combobox(qtbot: QtBot) -> None:
    # Arrange
    prediction_model: PredictionModel = PredictionModel()
    prediction_file_input: PredictionFileInput = PredictionFileInput(
        prediction_model, viewer=FakeViewer(), service=None
    )
    prediction_model.set_max_channels(6)

    # Act
    prediction_file_input._populate_input_channel_combobox()

    # Assert
    assert prediction_file_input._channel_select_dropdown.isEnabled()



from abc import ABC
from typing import Dict, Callable

from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.event import Event


class Publisher(ABC):
    """
    ABC publisher class
    """

    def __init__(self):
        # Map of Event -> Subscriber -> Handler function
        self._events_to_subscriber_handlers: Dict[
            Dict[Subscriber, Callable]
        ] = {event: dict() for event in [e.value for e in Event]}

    def dispatch(self, event: Event) -> None:
        """
        Dispatches an event to all subscribers
        """
        for _, handler in self._events_to_subscriber_handlers[
            event.value
        ].items():
            handler(event)

    def subscribe(self, event: Event, subscriber: Subscriber, handler) -> None:
        """
        subscribes a subscriber to this publisher
        """
        self._events_to_subscriber_handlers[event.value][subscriber] = handler

    def unsubscribe(self, event: Event, subscriber: Subscriber) -> None:
        """
        unsubscribes a subscriber from this publisher
        """
        if subscriber in self._events_to_subscriber_handlers[event.value]:
            del self._events_to_subscriber_handlers[event.value][subscriber]



from pathlib import Path

from allencell_ml_segmenter._tests.fakes.fake_channel_extraction import (
    FakeChannelExtractionThread,
)
from allencell_ml_segmenter.core.channel_extraction import (
    ChannelExtractionThread,
)
from allencell_ml_segmenter.core.i_extractor_factory import IExtractorFactory


class ExtractorFactory(IExtractorFactory):

    def create(self, img_path: Path) -> ChannelExtractionThread:
        return ChannelExtractionThread(img_path)


class FakeExtractorFactory(IExtractorFactory):
    def __init__(self, fake_value: int):
        super().__init__()
        self._fake_value = fake_value

    def create(self, img_path: Path) -> FakeChannelExtractionThread:
        return FakeChannelExtractionThread(self._fake_value)



import pytest
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.sample.sample_model import SampleModel
from allencell_ml_segmenter.sample.sample_state_widget import (
    TRAINING_NOT_RUNNING,
    TRAINING_RUNNING,
    SampleStateWidget,
)


@pytest.fixture
def sample_model() -> SampleModel:
    return SampleModel()


@pytest.fixture
def sample_state_widget(sample_model: SampleModel) -> SampleStateWidget:
    return SampleStateWidget(sample_model)


def test_handles_process_training_evt(
    qtbot: QtBot,
    sample_state_widget: SampleStateWidget,
    sample_model: SampleModel,
):
    # ARRANGE
    qtbot.addWidget(sample_state_widget)

    # ACT
    sample_model.set_process_running(True)

    # ASSERT
    assert sample_state_widget._label.text() == TRAINING_RUNNING


def test_handles_process_training_show_error_evt(
    qtbot: QtBot,
    sample_state_widget: SampleStateWidget,
    sample_model: SampleModel,
):
    # ARRANGE
    qtbot.addWidget(sample_state_widget)

    # ACT
    sample_model.set_error_message("Error")

    # ASSERT
    assert sample_state_widget._label.text() == "Error"

    # ACT
    sample_model.dispatch(Event.PROCESS_TRAINING_CLEAR_ERROR)

    # ASSERT
    assert sample_state_widget._label.text() == TRAINING_NOT_RUNNING



from typing import Dict, Union, Optional, List

from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
)
from allencell_ml_segmenter.utils.cuda_util import CUDAUtils


class CytoDLOverridesManager:
    """
    Class to generate overrides for cyto-dl based on user selections in app state
    """

    # TODO use a similar pattern for Predictions- willdo in another ticket
    # https://github.com/AllenCell/allencell-ml-segmenter/issues/234
    def __init__(
        self,
        experiments_model: ExperimentsModel,
        training_model: TrainingModel = None,
    ) -> None:
        self._experiments_model: ExperimentsModel = experiments_model
        self._training_model: Optional[TrainingModel] = training_model

    def get_training_overrides(
        self,
    ) -> Dict[str, Union[str, int, float, bool, Dict, List]]:
        # check to see if CytoOverridesManager was constructed with a training model
        if self._training_model is None:
            raise ValueError(
                "CytoOverridesManager must be constructed with a training model in order to get training overrides."
            )

        overrides_dict: Dict[str, Union[str, int, float, bool, Dict, List]] = (
            dict()
        )

        # Hardware override (required)
        if CUDAUtils.cuda_available():
            overrides_dict["trainer.accelerator"] = "gpu"
        else:
            overrides_dict["trainer.accelerator"] = "cpu"

        overrides_dict["data.num_workers"] = CUDAUtils.get_num_workers()

        # Spatial Dims (required)
        dims: int = self._training_model.get_spatial_dims()
        overrides_dict["spatial_dims"] = dims

        # Patch shape (required)
        patch_size: List[int] = self._training_model.get_patch_size()
        overrides_dict["data._aux.patch_shape"] = patch_size

        # Max Run
        # define max run (in epochs, required)
        current_epoch: int = self._experiments_model.get_current_epoch()
        current_epoch = current_epoch + 1 if current_epoch is not None else 0
        overrides_dict["trainer.max_epochs"] = (
            self._training_model.get_num_epochs() + current_epoch
        )
        # max run in time is also defined (optional)
        if self._training_model.use_max_time():
            # define max runtime (in hours)
            overrides_dict["trainer.max_time"] = {
                "minutes": self._training_model.get_max_time()
            }

        # Training input path (required)
        overrides_dict["data.path"] = str(
            self._training_model.get_images_directory()
        )

        # Commented out 5/23 brian.kim
        # We no longer support training from a old checkpoint- leaving this in if we want to re-enable this in the
        # future to continue training from an existing checkpoint.
        # if self._experiments_model.get_checkpoint() is not None:
        #     # We are going to continue training on an existing model
        #     overrides_dict["ckpt_path"] = str(
        #         self._experiments_model.get_model_checkpoints_path(
        #             self._experiments_model.get_experiment_name(),
        #             self._experiments_model.get_checkpoint(),
        #         )
        #     )

        # Filters/Model Size (required)
        overrides_dict["model._aux.filters"] = (
            self._training_model.get_model_size().value
        )

        # Channel Override
        overrides_dict["input_channel"] = (
            self._training_model.get_selected_channel(ImageType.RAW)
        )
        overrides_dict["target_col1_channel"] = (
            self._training_model.get_selected_channel(ImageType.SEG1)
        )
        overrides_dict["target_col2_channel"] = (
            self._training_model.get_selected_channel(ImageType.SEG2)
        )

        return overrides_dict



from watchdog.events import FileSystemEvent, FileSystemEventHandler
from typing import Callable, Set


class PredictionFolderEventHandler(FileSystemEventHandler):
    """
    A PredictionFolderEventHandler calls progress_callback upon creation of a file
    with an extension contained in PRED_FILE_EXTS. The argument to progress_callback
    will be how many files that fit these criteria have been created.
    """

    PRED_FILE_EXTS: Set[str] = {".tif", ".tiff"}

    def __init__(self, progress_callback: Callable):
        super().__init__()
        self._progress_callback: Callable = progress_callback
        self._num_pred_files_created = 0

    # override
    def on_created(self, event: FileSystemEvent) -> None:
        if any([event.src_path.endswith(ext) for ext in self.PRED_FILE_EXTS]):
            self._num_pred_files_created += 1
            self._progress_callback(self._num_pred_files_created)



from dataclasses import dataclass
from unittest.mock import Mock, patch
from typing import List
from pathlib import Path

import pytest
from qtpy.QtWidgets import QComboBox, QFileDialog
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.curation.input_view import CurationInputView
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    ImageType,
    CurationView,
)
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.main.main_model import MainModel


@dataclass
class TestEnvironment:
    model: CurationModel
    view: CurationInputView


@pytest.fixture
def test_env() -> TestEnvironment:
    model: CurationModel = CurationModel(FakeExperimentsModel(), MainModel())
    return TestEnvironment(model, CurationInputView(model))


MOCK_STR_PATH: str = "mock_path"
MOCK_DIR_PATHS: List[Path] = [Path("p1"), Path("p2")]

### UI State Tests ----------------------------------------------------------------------------------


def assert_combo_box_matches_channel_count(
    combo_box: QComboBox, count: int
) -> None:
    """
    Assert that the given combo box contains options 0 through count - 1 inclusive
    in ascending order.
    """
    expected_items = [str(x) for x in range(count)]
    assert combo_box.count() == len(expected_items)
    for i in range(combo_box.count()):
        assert combo_box.itemText(i) == expected_items[i]


@patch.multiple(
    QFileDialog,
    exec_=Mock(return_value=QFileDialog.Accepted),
    getExistingDirectory=Mock(return_value=MOCK_STR_PATH),
)
class TestsWithStubbedFileDialog:
    def test_raw_dir_selected(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Assert sanity check
        assert not test_env.view.raw_dir_stacked_spinner.is_spinning()
        # Act
        test_env.view.raw_directory_select.button.click()
        # Assert
        assert test_env.view.raw_dir_stacked_spinner.is_spinning()

    def test_seg1_dir_selected(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Assert sanity check
        assert not test_env.view.seg1_dir_stacked_spinner.is_spinning()
        # Act
        test_env.view.seg1_directory_select.button.click()
        # Assert
        assert test_env.view.seg1_dir_stacked_spinner.is_spinning()

    def test_seg2_dir_selected(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Assert sanity check
        assert not test_env.view.seg2_dir_stacked_spinner.is_spinning()
        # Act
        test_env.view.seg2_directory_select.button.click()
        # Assert
        assert test_env.view.seg2_dir_stacked_spinner.is_spinning()

    def test_raw_channel_count_set(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Arrange
        test_env.view.raw_directory_select.button.click()
        # Simulate the model's channel count being set
        count: int = 2
        combo_box: QComboBox = test_env.view.raw_image_channel_combo

        # Act
        test_env.model.set_channel_count(ImageType.RAW, count)

        # Assert
        assert not test_env.view.raw_dir_stacked_spinner.is_spinning()
        assert_combo_box_matches_channel_count(combo_box, count)

    def test_seg1_channel_count_set(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Arrange
        test_env.view.seg1_directory_select.button.click()
        # Simulate the model's channel count being set
        count: int = 3
        combo_box: QComboBox = test_env.view.seg1_image_channel_combo

        # Act
        test_env.model.set_channel_count(ImageType.SEG1, count)

        # Assert
        assert not test_env.view.seg1_dir_stacked_spinner.is_spinning()
        assert_combo_box_matches_channel_count(combo_box, count)

    def test_seg2_channel_count_set(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Arrange
        test_env.view.seg2_directory_select.button.click()
        # Simulate the model's channel count being set
        count: int = 4
        combo_box: QComboBox = test_env.view.seg2_image_channel_combo

        # Act
        test_env.model.set_channel_count(ImageType.SEG2, count)

        # Assert
        assert not test_env.view.seg2_dir_stacked_spinner.is_spinning()
        assert_combo_box_matches_channel_count(combo_box, count)

    def test_start_raw_selected(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Arrange
        test_env.view.raw_directory_select.button.click()
        test_env.model.set_channel_count(ImageType.RAW, 5)
        test_env.view.raw_image_channel_combo.setCurrentIndex(1)

        # Assert (sanity check)
        assert test_env.model.get_current_view() == CurationView.INPUT_VIEW

        # Act
        test_env.view.start_btn.click()

        # Assert
        # we expect the start button to fail since we haven't selected seg1 -- should still be input view
        assert test_env.model.get_current_view() == CurationView.INPUT_VIEW

    def test_start_raw_seg1_selected(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Arrange
        test_env.view.raw_directory_select.button.click()
        test_env.model.set_channel_count(ImageType.RAW, 5)
        test_env.model.set_image_directory_paths(ImageType.RAW, MOCK_DIR_PATHS)
        test_env.view.raw_image_channel_combo.setCurrentIndex(1)

        test_env.view.seg1_directory_select.button.click()
        test_env.model.set_channel_count(ImageType.SEG1, 5)
        test_env.model.set_image_directory_paths(
            ImageType.SEG1, MOCK_DIR_PATHS
        )
        test_env.view.seg1_image_channel_combo.setCurrentIndex(2)

        # Assert (sanity check)
        assert test_env.model.get_current_view() == CurationView.INPUT_VIEW

        # Act
        test_env.view.start_btn.click()

        # Assert
        # we expect the start button to work since we have at least raw and seg1
        assert test_env.model.get_current_view() == CurationView.MAIN_VIEW

    def test_start_all_selected(
        self, qtbot: QtBot, test_env: TestEnvironment
    ) -> None:
        # Arrange
        test_env.view.raw_directory_select.button.click()
        test_env.model.set_channel_count(ImageType.RAW, 5)
        test_env.model.set_image_directory_paths(ImageType.RAW, MOCK_DIR_PATHS)
        test_env.view.raw_image_channel_combo.setCurrentIndex(1)

        test_env.view.seg1_directory_select.button.click()
        test_env.model.set_channel_count(ImageType.SEG1, 5)
        test_env.model.set_image_directory_paths(
            ImageType.SEG1, MOCK_DIR_PATHS
        )
        test_env.view.seg1_image_channel_combo.setCurrentIndex(2)

        test_env.view.seg2_directory_select.button.click()
        test_env.model.set_channel_count(ImageType.SEG2, 5)
        test_env.model.set_image_directory_paths(
            ImageType.SEG2, MOCK_DIR_PATHS
        )
        test_env.view.seg2_image_channel_combo.setCurrentIndex(3)

        # Assert (sanity check)
        assert test_env.model.get_current_view() == CurationView.INPUT_VIEW

        # Act
        test_env.view.start_btn.click()

        # Assert
        # we expect the start button to work since we have all selected
        assert test_env.model.get_current_view() == CurationView.MAIN_VIEW


### View + Model Integration Tests ------------------------------------------------------------------------


def test_raw_channel_selected(qtbot: QtBot, test_env: TestEnvironment) -> None:
    # Arrange
    test_env.model.set_channel_count(ImageType.RAW, 3)
    combo_box: QComboBox = test_env.view.raw_image_channel_combo
    # Act / Assert
    assert test_env.model.get_selected_channel(ImageType.RAW) == 0
    combo_box.setCurrentIndex(0)
    assert test_env.model.get_selected_channel(ImageType.RAW) == 0
    combo_box.setCurrentIndex(1)
    assert test_env.model.get_selected_channel(ImageType.RAW) == 1
    combo_box.setCurrentIndex(2)
    assert test_env.model.get_selected_channel(ImageType.RAW) == 2


def test_seg1_channel_selected(
    qtbot: QtBot, test_env: TestEnvironment
) -> None:
    # Arrange
    test_env.model.set_channel_count(ImageType.SEG1, 3)
    combo_box: QComboBox = test_env.view.seg1_image_channel_combo
    # Act / Assert
    assert test_env.model.get_selected_channel(ImageType.SEG1) == 0
    combo_box.setCurrentIndex(0)
    assert test_env.model.get_selected_channel(ImageType.SEG1) == 0
    combo_box.setCurrentIndex(1)
    assert test_env.model.get_selected_channel(ImageType.SEG1) == 1
    combo_box.setCurrentIndex(2)
    assert test_env.model.get_selected_channel(ImageType.SEG1) == 2


def test_seg2_channel_selected(
    qtbot: QtBot, test_env: TestEnvironment
) -> None:
    # Arrange
    test_env.model.set_channel_count(ImageType.SEG2, 3)
    combo_box: QComboBox = test_env.view.seg2_image_channel_combo
    # Act / Assert
    assert test_env.model.get_selected_channel(ImageType.SEG2) == 0
    combo_box.setCurrentIndex(0)
    assert test_env.model.get_selected_channel(ImageType.SEG2) == 0
    combo_box.setCurrentIndex(1)
    assert test_env.model.get_selected_channel(ImageType.SEG2) == 1
    combo_box.setCurrentIndex(2)
    assert test_env.model.get_selected_channel(ImageType.SEG2) == 2



from pathlib import Path
from typing import List, Optional

from napari.utils.events import Event as NapariEvent
from napari.utils.notifications import show_warning
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QRadioButton,
    QComboBox,
    QGridLayout,
    QLabel,
    QFrame,
    QSizePolicy,
)

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.widgets.input_button_widget import (
    InputButton,
    FileInputMode,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
    PredictionInputMode,
)

from allencell_ml_segmenter.widgets.check_box_list_widget import (
    CheckBoxListWidget,
)

from allencell_ml_segmenter.main.viewer import Viewer
from allencell_ml_segmenter.prediction.service import ModelFileService
from allencell_ml_segmenter.curation.stacked_spinner import StackedSpinner


class PredictionFileInput(QWidget):
    """
    A widget containing file inputs for the input to a model prediction.
    """

    TOP_TEXT: str = "On-screen image(s)"
    BOTTOM_TEXT: str = "Image directory"

    def __init__(
        self, model: PredictionModel, viewer: Viewer, service: ModelFileService
    ):
        super().__init__()

        self._model: PredictionModel = model
        self._viewer: Viewer = viewer
        self._service: ModelFileService = service
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)

        # set up napari event listener for layer changes
        # this keeps the layer list in our UI updated as the layers are added/deleted from napari
        self._viewer.subscribe_layers_change_event(self._update_layer_list)

        frame: QFrame = QFrame()
        frame.setLayout(QVBoxLayout())

        frame.setObjectName("frame")

        title: LabelWithHint = LabelWithHint("Input image(s)")
        title.set_hint("Image(s) to apply the trained model on")
        title.setObjectName("title")

        self.layout().addWidget(title)
        self.layout().addWidget(frame)

        frame.layout().addWidget(QLabel("Select input image(s):"))

        # radiobox for images from napari
        horiz_layout: QHBoxLayout = QHBoxLayout()
        horiz_layout.setSpacing(0)

        self._radio_on_screen: QRadioButton = QRadioButton()
        self._radio_on_screen.toggled.connect(self._on_screen_slot)
        self._radio_on_screen.setObjectName("onScreen")

        horiz_layout.addWidget(self._radio_on_screen)

        question_label: LabelWithHint = LabelWithHint(
            PredictionFileInput.TOP_TEXT
        )
        question_label.set_hint("Image(s) already opened in napari")
        horiz_layout.addWidget(question_label)

        frame.layout().addLayout(horiz_layout)

        # list of available images on napari
        self._image_list: CheckBoxListWidget = CheckBoxListWidget()
        self._image_list.setEnabled(False)
        self._image_list.setObjectName("imageList")
        self._image_list.checkedSignal.connect(self._process_checked_signal)
        frame.layout().addWidget(self._image_list)

        # radiobox for images from directory
        horiz_layout = QHBoxLayout()
        horiz_layout.setSpacing(0)

        self._radio_directory: QRadioButton = QRadioButton()
        self._radio_directory.toggled.connect(self._from_directory_slot)
        self._radio_directory.setObjectName("radioDirectory")

        horiz_layout.addWidget(self._radio_directory)

        image_dir_layout: QVBoxLayout = QVBoxLayout()

        question_label = LabelWithHint(PredictionFileInput.BOTTOM_TEXT)
        question_label.set_hint(
            "Whole directory of image will be used as input. Prediction results will not be displayed in napari after prediction completion."
        )
        question_label.add_right_space(10)
        image_dir_layout.addWidget(question_label)

        horiz_layout.addLayout(image_dir_layout)

        horiz_layout.addStretch(5)

        self._browse_dir_edit: InputButton = InputButton(
            self._model,
            lambda dir: self._model.set_input_image_path(
                Path(dir), extract_channels=True
            ),
            "Select directory...",
            FileInputMode.DIRECTORY,
        )
        self.input_image_spinner: StackedSpinner = StackedSpinner(
            input_button=self._browse_dir_edit
        )
        self._browse_dir_edit.setEnabled(False)
        horiz_layout.addWidget(self.input_image_spinner)
        frame.layout().addLayout(horiz_layout)

        grid_layout: QGridLayout = QGridLayout()

        image_input_label: LabelWithHint = LabelWithHint(
            "Input image's channel"
        )
        image_input_label.set_hint(
            "Select which channel of the input image(s) to apply the trained model on"
        )

        self._channel_select_dropdown: QComboBox = QComboBox()

        self._channel_select_dropdown.setCurrentIndex(-1)
        self._channel_select_dropdown.currentIndexChanged.connect(
            self._model.set_image_input_channel_index
        )
        self._channel_select_dropdown.setEnabled(False)
        # Event to trigger combobox populate when we know the number of channels
        self._model.subscribe(
            Event.ACTION_PREDICTION_MAX_CHANNELS_SET,
            self,
            self._populate_input_channel_combobox,
        )
        # Event to set combobox text to 'loading' when we begin extracting channels
        self._model.subscribe(
            Event.ACTION_PREDICTION_EXTRACT_CHANNELS,
            self,
            self._set_input_channel_combobox_to_loading,
        )

        output_dir_label: LabelWithHint = LabelWithHint("Output directory")
        output_dir_label.set_hint(
            "Directory to store the prediction result(s)"
        )

        self._browse_output_edit: InputButton = InputButton(
            self._model,
            lambda dir: self._model.set_output_directory(dir),
            "Select directory...",
            FileInputMode.DIRECTORY,
        )

        grid_layout.addWidget(image_input_label, 0, 0)
        grid_layout.addWidget(self._channel_select_dropdown, 0, 1)

        grid_layout.addWidget(output_dir_label, 1, 0)
        grid_layout.addWidget(self._browse_output_edit, 1, 1)

        grid_layout.setColumnStretch(0, 1)
        grid_layout.setColumnStretch(1, 0)

        frame.layout().addLayout(grid_layout)

    def _on_screen_slot(self) -> None:
        """Prohibits usage of non-related input fields if top button is checked."""
        self._reset_channel_combobox()
        self._image_list.setEnabled(True)
        self._browse_dir_edit.setEnabled(False)
        self._update_layer_list()
        self._model.set_prediction_input_mode(
            PredictionInputMode.FROM_NAPARI_LAYERS
        )

    def _from_directory_slot(self) -> None:
        """Prohibits usage of non-related input fields if bottom button is checked."""
        self._reset_channel_combobox()
        self._image_list.setEnabled(False)
        self._browse_dir_edit.setEnabled(True)
        self._model.set_prediction_input_mode(PredictionInputMode.FROM_PATH)

    def _update_layer_list(self, event: Optional[NapariEvent] = None) -> None:
        self._image_list.clear()
        self._model.set_selected_paths([], extract_channels=False)
        self._reset_channel_combobox()
        for layer in self._viewer.get_layers():
            path_of_layer_image: str = layer.source.path
            if path_of_layer_image:
                self._image_list.add_item(layer.name)

    def _process_checked_signal(self, row: int, state: Qt.CheckState) -> None:
        if (
            self._model.get_prediction_input_mode()
            == PredictionInputMode.FROM_NAPARI_LAYERS
        ):
            selected_indices: List[int] = self._image_list.get_checked_rows()
            selected_paths: List[Path] = [
                Path(self._viewer.get_layers()[i].source.path)
                for i in selected_indices
            ]

            # this will preserve the invariant: the options in the dropdown will be equal
            # to the number of channels in at least one of the selected images (or empty if no images selected)
            if state == Qt.CheckState.Checked:
                if len(selected_paths) <= 10:
                    # only extract if it's the only one checked and it's just been checked;
                    # otherwise we assume they are checking images with same number of channels
                    self._model.set_selected_paths(
                        selected_paths,
                        extract_channels=len(selected_paths) == 1,
                    )
                else:
                    show_warning(
                        "Cannot predict on > 10 images from the viewer"
                    )
                    self._image_list.item(row).setCheckState(
                        Qt.CheckState.Unchecked
                    )
            elif state == Qt.CheckState.Unchecked:
                # could have unselected the img we got channels from originally, so need to re-extract
                # as long as there are still some images selected
                if len(selected_indices) == 0:
                    self._service.stop_channel_extraction()  # stop so combobox doesn't reset after thread is finished
                    self._reset_channel_combobox()

                self._model.set_selected_paths(
                    selected_paths, extract_channels=len(selected_paths) > 0
                )

    def _reset_channel_combobox(self) -> None:
        self._channel_select_dropdown.clear()
        self._channel_select_dropdown.setPlaceholderText("")
        self._channel_select_dropdown.setCurrentIndex(-1)
        self._channel_select_dropdown.setEnabled(False)

    def _set_input_channel_combobox_to_loading(
        self, event: Event = None
    ) -> None:
        self.input_image_spinner.start()
        self._channel_select_dropdown.clear()
        self._channel_select_dropdown.setPlaceholderText("loading channels...")
        self._channel_select_dropdown.setCurrentIndex(-1)
        self._channel_select_dropdown.setEnabled(False)

    def _populate_input_channel_combobox(self, event: Event = None) -> None:
        self.input_image_spinner.stop()
        channels_in_image: Optional[int] = self._model.get_max_channels()
        self._reset_channel_combobox()
        if channels_in_image is not None and channels_in_image > 0:
            values_range: List[str] = [
                str(i) for i in range(self._model.get_max_channels())
            ]
            self._channel_select_dropdown.setPlaceholderText(
                "select a channel index"
            )

            self._channel_select_dropdown.addItems(values_range)
            self._channel_select_dropdown.setEnabled(True)
        else:
            self._channel_select_dropdown.setPlaceholderText(
                "no channels to select"
            )



from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    CurationRecord,
    ImageType,
    CurationView,
)
from allencell_ml_segmenter.core.image_data_extractor import (
    IImageDataExtractor,
    AICSImageDataExtractor,
    ImageData,
)
from allencell_ml_segmenter.core.task_executor import (
    ITaskExecutor,
    NapariThreadTaskExecutor,
)
from allencell_ml_segmenter.utils.file_utils import FileUtils
from allencell_ml_segmenter.utils.file_writer import IFileWriter, FileWriter
from allencell_ml_segmenter.main.main_model import MIN_DATASET_SIZE

from pathlib import Path
from qtpy.QtCore import QObject
from typing import List, Optional, Callable
from copy import deepcopy
from collections import namedtuple


DirectoryData = namedtuple("DirectoryData", ["fpaths", "channels"])


# Important note: we do not want to access the model in any of the threads because model state may change
# while thread is executing. So, opt to copy/pass in all relevant model state to threads
class CurationService(QObject):

    def __init__(
        self,
        curation_model: CurationModel,
        experiments_model: IExperimentsModel,
        img_data_extractor: IImageDataExtractor = AICSImageDataExtractor.global_instance(),
        task_executor: ITaskExecutor = NapariThreadTaskExecutor.global_instance(),
        file_writer: IFileWriter = FileWriter.global_instance(),
    ) -> None:
        super().__init__()
        self._curation_model: CurationModel = curation_model
        self._experiments_model: IExperimentsModel = experiments_model
        self._img_data_extractor: IImageDataExtractor = img_data_extractor
        self._task_executor: ITaskExecutor = task_executor
        self._file_writer: IFileWriter = file_writer
        self._file_utils: FileUtils = FileUtils(file_writer)

        self._curation_model.image_directory_set.connect(
            self._on_image_dir_set
        )
        self._curation_model.cursor_moved.connect(self._on_cursor_moved)
        self._curation_model.save_to_disk_requested.connect(
            self._on_save_to_disk
        )

    def _get_dir_data(self, dir: Path) -> DirectoryData:
        files: List[Path] = (
            self._file_utils.get_all_files_in_dir_ignore_hidden(dir)
        )

        if len(files) < MIN_DATASET_SIZE:
            raise RuntimeError(
                f"Curation requires at least {MIN_DATASET_SIZE} images and their segmentations"
            )

        img_data: ImageData = self._img_data_extractor.extract_image_data(
            files[0], np_data=False
        )
        return DirectoryData(files, img_data.channels)

    def _on_dir_data_extracted(
        self, img_type: ImageType, dir_data: DirectoryData
    ) -> None:
        self._curation_model.set_image_directory_paths(
            img_type, dir_data.fpaths
        )
        self._curation_model.set_channel_count(img_type, dir_data.channels)

    def _on_dir_data_errored(self, img_type: ImageType, e: Exception) -> None:
        self._curation_model.set_channel_count(img_type, 0)
        raise e

    def _on_image_dir_set(self, img_type: ImageType) -> None:
        # paths are immutable, so no need to explicitly copy
        dir: Path = self._curation_model.get_image_directory(img_type)
        self._task_executor.exec(
            lambda: self._get_dir_data(dir),
            on_return=lambda dir_data: self._on_dir_data_extracted(
                img_type, dir_data
            ),
            on_error=lambda e: self._on_dir_data_errored(img_type, e),
        )

    def _extract_images(
        self,
        img_idx: int,
        setter_fn: Callable[[ImageType, ImageData], None],
        err_str: str,
    ) -> None:
        """
        Uses TaskExecutor to extract data for images at :param img_idx:, saves extracted data using
        :param setter_fn: to set model state. Provides :param err_str: ('curr' or 'next') to the error
        handler for additional debugging info.
        """
        raw_paths: List[Path] = self._curation_model.get_image_directory_paths(
            ImageType.RAW
        )
        seg1_paths: List[Path] = (
            self._curation_model.get_image_directory_paths(ImageType.SEG1)
        )
        seg2_paths: Optional[List[Path]] = (
            self._curation_model.get_image_directory_paths(ImageType.SEG2)
        )

        raw_channel: int = self._curation_model.get_selected_channel(
            ImageType.RAW
        )
        seg1_channel: int = self._curation_model.get_selected_channel(
            ImageType.SEG1
        )
        seg2_channel: Optional[int] = (
            self._curation_model.get_selected_channel(ImageType.SEG2)
        )

        self._task_executor.exec(
            lambda: self._img_data_extractor.extract_image_data(
                raw_paths[img_idx],
                channel=raw_channel,
            ),
            on_return=lambda img_data: setter_fn(ImageType.RAW, img_data),
            on_error=lambda e: self._on_cursor_moved_error(
                ImageType.RAW, err_str, e
            ),
        )
        self._task_executor.exec(
            lambda: self._img_data_extractor.extract_image_data(
                seg1_paths[img_idx], channel=seg1_channel, seg=1
            ),
            on_return=lambda img_data: setter_fn(ImageType.SEG1, img_data),
            on_error=lambda e: self._on_cursor_moved_error(
                ImageType.SEG1, err_str, e
            ),
        )
        if seg2_paths is not None:
            self._task_executor.exec(
                lambda: self._img_data_extractor.extract_image_data(
                    seg2_paths[img_idx], channel=seg2_channel, seg=2
                ),
                on_return=lambda img_data: setter_fn(ImageType.SEG2, img_data),
                on_error=lambda e: self._on_cursor_moved_error(
                    ImageType.SEG2, err_str, e
                ),
            )

    def _on_cursor_moved_error(
        self, img_type: ImageType, curr_or_next: str, e: Exception
    ) -> None:
        raise RuntimeError(
            f"There was a problem loading {img_type} for the {curr_or_next} images"
        )

    def _on_cursor_moved(self) -> None:
        cursor: int = self._curation_model.get_curr_image_index()

        if self._curation_model.is_waiting_for_curr_images():
            # start extraction tasks for curr images (paths at cursor)
            self._extract_images(
                cursor, self._curation_model.set_curr_image_data, "curr"
            )

        if self._curation_model.is_waiting_for_next_images():
            # start extraction tasks for next images (paths at cursor + 1)
            self._extract_images(
                cursor + 1, self._curation_model.set_next_image_data, "next"
            )

    def _on_save_to_disk_error(self, err: Exception) -> None:
        self._curation_model.set_curation_record_saved_to_disk(False)
        raise err

    def _on_save_to_disk(self) -> None:
        record: List[CurationRecord] = deepcopy(
            self._curation_model.get_curation_record()
        )
        csv_path: Path = self._curation_model.get_csv_path()
        save_path: Path = self._curation_model.get_save_masks_path()
        self._task_executor.exec(
            lambda: self._file_utils.write_curation_record(
                record, csv_path, save_path
            ),
            on_finish=lambda: self._curation_model.set_curation_record_saved_to_disk(
                True
            ),
            on_error=self._on_save_to_disk_error,
        )



# this file is intended to be called by a github workflow (.github/workflows/publish_to_pypi.yaml)
# it makes decisions based on the current version and the component specified for bumping,
# which the workflow cannot do
import subprocess
import sys
import toml  # type: ignore


def main() -> None:
    if len(sys.argv) < 2:
        raise ValueError("No component specified for bumping version")

    component: str = sys.argv[1].lower()
    valid_options: set[str] = {"major", "minor", "patch", "dev", "post"}

    if component not in valid_options:
        raise ValueError(f"Component must be one of {valid_options}")

    version: str = toml.load("pyproject.toml")["project"]["version"]
    version_components: list[str] = version.split(".")

    update_output: subprocess.CompletedProcess
    # 4 components means we currently have a dev or post version
    if len(version_components) == 4 and version_components[-1].startswith(
        "dev"
    ):
        if component == "dev":
            # increment the dev tag (e.g. 1.0.0.dev0 -> 1.0.0.dev1)
            update_output = subprocess.run(
                ["bumpver", "update", "--tag-num", "-n"]
            )
        elif component == "patch":
            # finalize the patch by removing dev tag (e.g. 1.0.0.dev1 -> 1.0.0)
            update_output = subprocess.run(
                ["bumpver", "update", "--tag=final", "-n"]
            )
        else:
            raise ValueError(
                "Cannot update major or minor version while dev version is current"
            )
    elif len(version_components) == 4:  # current version must be post
        if component == "post":
            update_output = subprocess.run(
                ["bumpver", "update", "--tag-num", "-n"]
            )
        else:
            raise ValueError("Cannot change post version to standard version")
    elif len(version_components) == 3:
        if component == "dev":
            # increment patch and begin at dev0 (e.g. 1.0.0 -> 1.0.1.dev0)
            update_output = subprocess.run(
                ["bumpver", "update", "--patch", "--tag=dev", "-n"]
            )
        elif component == "post":
            update_output = subprocess.run(
                ["bumpver", "update", "--tag=post", "-n"]
            )
        else:
            update_output = subprocess.run(
                ["bumpver", "update", f"--{component}", "-n"]
            )

    else:
        raise ValueError(
            f"Unknown version format: {version}. Expected MAJOR.MINOR.PATCH[.PYTAGNUM]"
        )

    if update_output.returncode != 0:
        raise RuntimeError(
            f"bumpver exited with code {update_output.returncode}"
        )


if __name__ == "__main__":
    main()



import pytest
from pathlib import Path
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.main.main_model import MainModel

from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    TrainingType,
    ImageType,
)
from unittest.mock import Mock


@pytest.fixture
def training_model() -> TrainingModel:
    """
    Returns a TrainingModel instance for testing.
    """
    return TrainingModel(MainModel(), FakeExperimentsModel())


def test_get_experiment_type(training_model: TrainingModel) -> None:
    """
    Tests that get_experiment_type returns the correct experiment type.
    """
    # ASSERT
    assert training_model.get_experiment_type() is None

    # ARRANGE
    experiment: TrainingType = TrainingType("segmentation")
    training_model.set_experiment_type(experiment.value)

    # ACT/ASSERT
    assert training_model.get_experiment_type() == experiment.value


def test_set_experiment_type(training_model: TrainingModel) -> None:
    """
    Tests that set_experiment_type sets the correct experiment type.
    """
    training_type_str = "segmentation"

    # ACT
    training_model.set_experiment_type(training_type_str)

    # ASSERT
    assert training_model.get_experiment_type() == training_type_str


def test_set_experiment_type_invalid_experiment(
    training_model: TrainingModel,
) -> None:
    """
    Tests that set_experiment_type raises a ValueError when given an invalid experiment type.
    """
    # ACT
    with pytest.raises(ValueError):
        training_model.set_experiment_type("invalid_experiment")


def test_get_image_dims(training_model: TrainingModel) -> None:
    """
    Tests that get_image_dims returns the correct image dimensions.
    """
    # ASSERT
    assert training_model.get_spatial_dims() is None

    # ARRANGE
    training_model._spatial_dims = 2

    # ACT/ASSERT
    assert training_model.get_spatial_dims() == 2

    # ARRANGE
    training_model._spatial_dims = 3

    # ACT/ASSERT
    assert training_model.get_spatial_dims() == 3


def test_get_num_epochs(training_model: TrainingModel) -> None:
    """
    Tests that get_max_epoch returns the correct max epoch.
    """
    # ASSERT
    assert training_model.get_num_epochs() is None

    # ARRANGE
    training_model._num_epochs = 100

    # ACT/ASSERT
    assert training_model.get_num_epochs() == 100


def test_set_num_epochs(training_model: TrainingModel) -> None:
    """
    Tests that set_max_epoch sets the correct max epoch.
    """
    # ACT
    training_model.set_num_epochs(100)

    # ASSERT
    assert training_model._num_epochs == 100


def test_get_images_directory(training_model: TrainingModel) -> None:
    """
    Tests that get_images_directory returns the correct images directory.
    """
    # ASSERT
    assert training_model.get_images_directory() is None

    # ARRANGE
    path: Path = Path("/path/to/images")
    training_model._images_directory = path

    # ACT/ASSERT
    assert training_model.get_images_directory() == path


def test_set_images_directory(training_model: TrainingModel) -> None:
    """
    Tests that set_images_directory sets the correct images directory.
    """
    # ARRANGE
    img_dir_listener: Mock = Mock()
    path: Path = Path("/path/to/images")
    training_model.signals.images_directory_set.connect(img_dir_listener)

    # ACT
    training_model.set_images_directory(path)

    # ASSERT
    assert training_model._images_directory == path
    img_dir_listener.assert_called_once()


def test_selected_channel(training_model: TrainingModel) -> None:
    """
    Tests that get_channel_index returns the correct channel index.
    """
    # ASSERT
    assert training_model.get_selected_channel(ImageType.RAW) is None
    assert training_model.get_selected_channel(ImageType.SEG1) is None
    assert training_model.get_selected_channel(ImageType.SEG2) is None

    # ACT
    training_model.set_selected_channel(ImageType.RAW, 1)
    training_model.set_selected_channel(ImageType.SEG1, 2)
    training_model.set_selected_channel(ImageType.SEG2, 3)

    # ACT/ASSERT
    assert training_model.get_selected_channel(ImageType.RAW) == 1
    assert training_model.get_selected_channel(ImageType.SEG1) == 2
    assert training_model.get_selected_channel(ImageType.SEG2) == 3


def test_num_channels(training_model: TrainingModel) -> None:
    # ASSSERT
    assert training_model.get_num_channels(ImageType.RAW) is None
    assert training_model.get_num_channels(ImageType.SEG1) is None
    assert training_model.get_num_channels(ImageType.SEG2) is None

    # ARRANGE
    num_channel_listener: Mock = Mock()
    training_model.signals.num_channels_set.connect(num_channel_listener)

    # ACT
    training_model.set_all_num_channels(
        {
            ImageType.RAW: 1,
            ImageType.SEG1: 2,
            ImageType.SEG2: 3,
        }
    )

    # ASSERT
    assert training_model.get_num_channels(ImageType.RAW) == 1
    assert training_model.get_num_channels(ImageType.SEG1) == 2
    assert training_model.get_num_channels(ImageType.SEG2) == 3
    num_channel_listener.assert_called_once()


def test_get_patch_size(training_model: TrainingModel) -> None:
    """
    Tests that get_patch_size returns the correct patch size.
    """
    # ASSERT
    assert training_model.get_patch_size() is None
    expected_patch = [4, 8]

    # ARRANGE
    training_model._patch_size = expected_patch

    # ACT/ASSERT
    assert training_model.get_patch_size() == expected_patch


def test_set_patch_size(training_model: TrainingModel) -> None:
    """
    Tests that set_patch_size sets the correct patch size.
    """
    # expected
    patch_2d = [4, 8]
    patch_3d = [1, 2, 3]

    # ACT
    training_model.set_patch_size(patch_2d)

    # ASSERT
    assert training_model._patch_size == patch_2d

    # ACT
    training_model.set_patch_size(patch_3d)

    # ASSERT
    assert training_model._patch_size == patch_3d


def test_set_patch_size_invalid(training_model: TrainingModel) -> None:
    """
    Tests that set_patch_size throws an error if an invalid patch size is provided
    """

    # ACT/ASSERT
    with pytest.raises(ValueError):
        training_model.set_patch_size([1])

    with pytest.raises(ValueError):
        training_model.set_patch_size([1, 2, 3, 4])

    with pytest.raises(TypeError):
        training_model.set_patch_size(2)

    with pytest.raises(TypeError):
        training_model.set_patch_size(None)


def test_get_max_time(training_model: TrainingModel) -> None:
    """
    Tests that get_max_time returns the correct max time.
    """
    # ASSERT
    assert training_model.get_max_time() is None

    # ARRANGE
    training_model._max_time = 100

    # ACT/ASSERT
    assert training_model.get_max_time() == 100


def test_set_max_time(training_model: TrainingModel) -> None:
    """
    Tests that set_max_time sets the correct max time.
    """
    # ACT
    training_model.set_max_time(100)

    # ASSERT
    assert training_model._max_time == 100


def test_get_config_dir(training_model: TrainingModel) -> None:
    """
    Tests that get_config_dir returns the correct config directory.
    """
    # ASSERT
    assert training_model.get_config_dir() is None

    # ARRANGE
    path: Path = Path("/path/to/config")
    training_model._config_dir = path

    # ACT/ASSERT
    assert training_model.get_config_dir() == path


def test_set_config_dir(training_model: TrainingModel) -> None:
    """
    Tests that set_config_dir sets the correct config directory.
    """
    # ARRANGE
    path: Path = Path("/path/to/config")

    # ACT
    training_model.set_config_dir(path)

    # ASSERT
    assert training_model._config_dir == path



from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
)

from allencell_ml_segmenter.sample.sample_model import SampleModel
from allencell_ml_segmenter.core.aics_widget import AicsWidget
from qtpy.QtWidgets import QPushButton


class SampleSelectFilesWidget(AicsWidget):
    """
    Sets training files in the model.
    """

    def __init__(self, model: SampleModel):
        super().__init__()
        self._model = model

        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._btn = QPushButton("Select Files")
        self._btn.clicked.connect(
            lambda: self._model.set_training_input_files(
                [
                    "/path/to/file1",
                    "/path/to/file2",
                    "/path/to/file3",
                    "/path/to/file4",
                    "/path/to/file5",
                ]
            )
        )

        self.layout().addWidget(self._btn)

    def handle_event(self, event):
        pass



from pathlib import Path
from typing import Callable, Optional

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.training.training_model import TrainingModel


class FakeChannelsReady:
    def __init__(self):
        self.connected: Callable = None

    def connect(self, connect: Callable):
        self.connected = connect


class FakeChannelExtractionThread:
    """
    Fake of Channel Extraction Thread
    """

    def __init__(self, fake_value: int):
        self.channels_ready = FakeChannelsReady()
        self.started: bool = False
        self._fake_value = fake_value

    def start(self):
        self.started = True
        self.channels_ready.connected(self._fake_value)



from typing import Dict

from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
    QStackedWidget,
)

from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.view import View, MainWindow
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    CurationView,
)
from allencell_ml_segmenter.curation.input_view import CurationInputView
from allencell_ml_segmenter.curation.main_view import CurationMainView
import napari


class CurationUiMeta(type(QStackedWidget), type(Subscriber)):
    """
    Metaclass for MainWidget

    """

    pass


class CurationWidget(
    MainWindow, QStackedWidget, Subscriber, metaclass=CurationUiMeta
):
    def __init__(
        self,
        viewer: napari.Viewer,
        curation_model: CurationModel,
    ) -> None:
        super().__init__()
        self.viewer: napari.Viewer = viewer
        self.view_to_index: Dict[View, int] = dict()
        self.curation_model: CurationModel = curation_model
        # basic styling
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.curation_input_view: CurationInputView = CurationInputView(
            self.curation_model
        )
        self.initialize_view(self.curation_input_view)

        self.curation_main_view: CurationMainView = CurationMainView(
            self.curation_model, self.viewer
        )
        self.initialize_view(self.curation_main_view)

        self.set_view(self.curation_input_view)
        self.curation_model.current_view_changed.connect(self._on_view_changed)

    def _on_view_changed(self) -> None:
        if self.curation_model.get_current_view() == CurationView.INPUT_VIEW:
            self.set_view(self.curation_input_view)
        elif self.curation_model.get_current_view() == CurationView.MAIN_VIEW:
            self.set_view(self.curation_main_view)

    def set_view(self, view: View) -> None:
        """
        Set the current views, must be initialized first
        """
        self.setCurrentIndex(self.view_to_index[view])

    # NOTE: this is mostly just a testing convenience function
    def get_view(self) -> CurationView:
        if self.currentWidget() == self.curation_input_view:
            return CurationView.INPUT_VIEW
        else:
            return CurationView.MAIN_VIEW

    def initialize_view(self, view: View) -> None:
        """
        Initialize views. This is necessary because QStackedWidget requires all child widgets to be added
        """
        # QStackedWidget count method keeps track of how many child widgets have been added
        self.view_to_index[view] = self.count()
        self.addWidget(view)

    def focus_changed(self):
        # if we haven't finished curation, then reload current images
        if (
            self.currentWidget() == self.curation_main_view
            and not self.curation_model.get_image_loading_stopped()
        ):
            self.curation_main_view.add_curr_images_to_widget()



import pytest
from unittest.mock import Mock
from allencell_ml_segmenter.core.aics_widget import AicsWidget
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.core.publisher import Event
from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber


@pytest.fixture
def main_model() -> MainModel:
    """
    Fixture for MainModel testing.
    """
    return MainModel()


@pytest.fixture
def fake_subscriber() -> FakeSubscriber:
    """
    Fixture for FakeSubscriber testing.
    """
    return FakeSubscriber()


def test_get_current_view(main_model: MainModel) -> None:
    """
    Tests that the current view is correctly retrievable.
    """
    # ASSERT
    assert main_model.get_current_view() is None

    # ARRANGE
    mock_view: AicsWidget = Mock(spec=AicsWidget)
    main_model._current_view = mock_view

    # ACT/ASSERT
    assert main_model.get_current_view() == mock_view


def test_set_new_model(
    main_model: MainModel, fake_subscriber: FakeSubscriber
) -> None:
    """
    Tests that a new model can be set.
    """
    # ARRANGE
    main_model.subscribe(
        Event.ACTION_NEW_MODEL, fake_subscriber, fake_subscriber.handle
    )

    # ACT
    main_model.set_new_model(True)

    # ASSERT
    assert fake_subscriber.was_handled(Event.ACTION_NEW_MODEL)


def test_set_current_view(
    main_model: MainModel, fake_subscriber: FakeSubscriber
) -> None:
    """
    Tests that the current view is correctly settable.
    """
    # ARRANGE
    mock_view: AicsWidget = Mock(spec=AicsWidget)

    main_model.subscribe(
        Event.ACTION_CHANGE_VIEW, fake_subscriber, fake_subscriber.handle
    )

    # ACT
    main_model.set_current_view(mock_view)

    # ASSERT
    assert fake_subscriber.was_handled(Event.ACTION_CHANGE_VIEW)



from pathlib import Path
import pytest
import responses

import allencell_ml_segmenter
from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.utils.s3.s3_bucket_constants import PROD_BUCKET
from allencell_ml_segmenter.widgets.model_download_dialog import (
    ModelDownloadDialog,
)
from allencell_ml_segmenter._tests.utils.s3.s3_response_fixtures import (
    s3_response_listobjectv2_contents_two_models,
)


@pytest.fixture
@responses.activate
def model_download_dialog(
    s3_response_listobjectv2_contents_two_models: str,
) -> ModelDownloadDialog:
    fake_settings: FakeUserSettings = FakeUserSettings()
    fake_settings.set_user_experiments_path(
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "output_test_folder"
    )
    exp_model: ExperimentsModel = ExperimentsModel(fake_settings)

    # add fake xml response that is returned when we make a request to the test_url with the list-type=2 param
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{PROD_BUCKET}?list-type=2",
            "body": s3_response_listobjectv2_contents_two_models,
            "status": 200,
            "content_type": "application/xml",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    return ModelDownloadDialog(None, exp_model)


def test_model_download_dialog_init(
    model_download_dialog: ModelDownloadDialog,
):
    # ASSERT
    assert (
        model_download_dialog._model_select_dropdown.count() == 2
    )  # two items added to combobox



from allencell_ml_segmenter.training.patch_size_validator import (
    PatchSizeValidator,
)
from qtpy.QtGui import QValidator
import pytest


@pytest.fixture
def validator() -> PatchSizeValidator:
    return PatchSizeValidator()


def test_validate_invalid(validator: PatchSizeValidator) -> None:
    assert validator.validate("0", 0)[0] == QValidator.State.Invalid
    assert validator.validate("-10", 0)[0] == QValidator.State.Invalid
    assert validator.validate("f", 0)[0] == QValidator.State.Invalid
    assert validator.validate("some string", 3)[0] == QValidator.State.Invalid
    assert validator.validate("+", 0)[0] == QValidator.State.Invalid
    assert validator.validate("0x", 0)[0] == QValidator.State.Invalid
    assert validator.validate("0b", 0)[0] == QValidator.State.Invalid
    assert validator.validate("0o", 0)[0] == QValidator.State.Invalid


def test_validate_intermediate(validator: PatchSizeValidator) -> None:
    assert validator.validate("1", 0)[0] == QValidator.State.Intermediate
    assert validator.validate("2", 0)[0] == QValidator.State.Intermediate
    assert validator.validate("3", 0)[0] == QValidator.State.Intermediate
    assert validator.validate("22", 0)[0] == QValidator.State.Intermediate
    assert validator.validate("109", 0)[0] == QValidator.State.Intermediate


def test_validate_acceptable(validator: PatchSizeValidator) -> None:
    assert validator.validate("4", 0)[0] == QValidator.State.Acceptable
    assert validator.validate("8", 0)[0] == QValidator.State.Acceptable
    assert validator.validate("64", 0)[0] == QValidator.State.Acceptable
    assert validator.validate("72", 0)[0] == QValidator.State.Acceptable
    assert validator.validate("128", 0)[0] == QValidator.State.Acceptable


def test_fixup(validator: PatchSizeValidator) -> None:
    assert validator.fixup("1") == "4"
    assert validator.fixup("3") == "4"
    assert validator.fixup("5") == "4"
    assert validator.fixup("9") == "8"
    assert validator.fixup("11") == "8"
    assert validator.fixup("131") == "128"



from enum import Enum
from pathlib import Path
from typing import List, Optional

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.publisher import Publisher


class PredictionInputMode(Enum):
    FROM_PATH = "from_path"
    FROM_NAPARI_LAYERS = "from_napari_layers"


class PredictionModel(Publisher):
    """
    Stores state relevant to prediction processes.
    """

    def __init__(self):
        super().__init__()

        # state related to PredictionFileInput
        self.config_name: str = None
        self.config_dir: Path = None
        self._input_image_path: Path = None
        self._image_input_channel_index: int = None
        self._input_mode: PredictionInputMode = None
        self._output_directory: Path = None
        self._selected_paths: List[Path] = None
        self._max_channels: int = None

        # state related to ModelInputWidget
        self._preprocessing_method: str = None
        self._postprocessing_method: str = None
        self._postprocessing_simple_threshold: float = None
        self._postprocessing_auto_threshold: str = None

        # This is initialized as None, and set when the input data is processed during pre-processing
        # If it is none after Event.ACTION_PREDICTION_SETUP is dispatched, the csv for
        # prediction was not generated and prediction cannot continue.
        self.total_num_images: Optional[int] = None

    def get_input_image_path(self) -> Path:
        """
        Gets list of paths to input images.
        """
        return self._input_image_path

    def set_input_image_path(
        self, path: Path, extract_channels: bool = False
    ) -> None:
        """
        Sets list of paths to input images.
        """
        self._input_image_path = path
        if extract_channels:
            # This will extract and set number of channels
            self.dispatch(Event.ACTION_PREDICTION_EXTRACT_CHANNELS)

    def get_image_input_channel_index(self) -> int:
        """
        Gets path to output directory.
        """
        return self._image_input_channel_index

    def set_image_input_channel_index(self, idx: int) -> None:
        """
        Sets path to output directory.
        """
        self._image_input_channel_index = idx

    def get_output_directory(self) -> Path:
        """
        Gets path to output directory.
        """
        return self._output_directory

    def get_output_seg_directory(self) -> Path:
        """
        Gets path to where segmentation predictions are stored.
        """
        return self._output_directory / "target"

    def set_output_directory(self, dir: Path) -> None:
        """
        Sets path to output directory.
        """
        self._output_directory = dir

    def get_model_path(self) -> Path:
        """
        Gets path to model.
        """
        return self._model_path

    def set_model_path(self, path: Path) -> None:
        """
        Sets path to model.
        """
        self._model_path = path
        self.dispatch(Event.ACTION_PREDICTION_MODEL_FILE)

    def get_preprocessing_method(self) -> str:
        """
        Gets preprocessing method associated with currently-selected model.
        """
        return self._preprocessing_method

    def set_preprocessing_method(self, method: str) -> None:
        """
        Sets preprocessing method associated with model after the service parses the file.
        """
        self._preprocessing_method = method
        self.dispatch(Event.ACTION_PREDICTION_PREPROCESSING_METHOD)

    def get_postprocessing_method(self) -> str:
        """
        Gets postprocessing method selected by user.
        """
        return self._postprocessing_method

    def set_postprocessing_method(self, method: str) -> None:
        """
        Sets postprocessing method selected by user.
        """
        self._postprocessing_method = method
        self.dispatch(Event.ACTION_PREDICTION_POSTPROCESSING_METHOD)

    def get_postprocessing_simple_threshold(self) -> float:
        """
        Gets simple threshold selected by user.
        """
        return self._postprocessing_simple_threshold

    def set_postprocessing_simple_threshold(self, threshold: float) -> None:
        """
        Sets simple threshold selected by user from the slider.
        """
        self._postprocessing_simple_threshold = threshold
        self.dispatch(Event.ACTION_PREDICTION_POSTPROCESSING_SIMPLE_THRESHOLD)

    def get_postprocessing_auto_threshold(self) -> str:
        """
        Gets auto threshold selected by user.
        """
        return self._postprocessing_auto_threshold

    def set_postprocessing_auto_threshold(self, threshold: str) -> None:
        """
        Sets auto threshold selected by user.
        """
        self._postprocessing_auto_threshold = threshold
        self.dispatch(Event.ACTION_PREDICTION_POSTPROCESSING_AUTO_THRESHOLD)

    def set_config_dir(self, config_dir: Path) -> None:
        self.config_dir = config_dir

    def get_config_dir(self) -> Path:
        return self.config_dir

    def set_config_name(self, config_name: str) -> None:
        self.config_name = config_name

    def get_config_name(self) -> str:
        return self.config_name

    def set_prediction_input_mode(self, mode: PredictionInputMode) -> None:
        self._input_mode = mode

    def get_prediction_input_mode(self) -> PredictionInputMode:
        return self._input_mode

    def set_selected_paths(
        self, paths: List[Path], extract_channels: bool = False
    ) -> None:
        self._selected_paths = paths
        if extract_channels:
            self.dispatch(Event.ACTION_PREDICTION_EXTRACT_CHANNELS)

    def get_selected_paths(self) -> List[Path]:
        return self._selected_paths

    def dispatch_prediction(self) -> None:
        """
        Dispatches an event to start prediction
        """
        # Shoots off a prediction run
        self.dispatch(Event.PROCESS_PREDICTION)

    def dispatch_prediction_get_image_paths_from_napari(self) -> None:
        # Does some pre-configuring if needed for prediction runs
        self.dispatch(Event.ACTION_PREDICTION_GET_IMAGE_PATHS_FROM_NAPARI)

    def dispatch_prediction_setup(self) -> None:
        # Does some pre-configuring if needed for prediction runs
        self.dispatch(Event.ACTION_PREDICTION_SETUP)

    def set_max_channels(self, max: int) -> None:
        self._max_channels = max
        # this will enable the combobox
        self.dispatch(Event.ACTION_PREDICTION_MAX_CHANNELS_SET)

    def get_max_channels(self) -> int:
        return self._max_channels

    def set_total_num_images(self, total: Optional[int]):
        self.total_num_images = total

    def get_total_num_images(self) -> Optional[int]:
        return self.total_num_images



from allencell_ml_segmenter.core.q_runnable_manager import IQRunnableManager
from qtpy.QtCore import QRunnable


class SynchroQRunnableManager(IQRunnableManager):
    """
    Executes QRunnables on the calling thread. Used for testing
    """

    _instance = None

    def run(self, runnable: QRunnable) -> None:
        runnable.run()

    @classmethod
    def global_instance(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



from pathlib import Path
from qtpy.QtCore import QSettings
from qtpy.QtWidgets import QFileDialog
from qtpy.QtWidgets import QMessageBox
from qtpy.QtWidgets import QWidget

from allencell_ml_segmenter.config.i_user_settings import IUserSettings

CYTO_DL_HOME_PATH = "/Users/chrishu/dev/code/test2/cyto-dl"
EXPERIMENTS_HOME_KEY = "experimentshome"


class UserSettings(IUserSettings):
    def __init__(
        self, settings: QSettings = QSettings("AICS", "Segmenter ML")
    ):
        self.settings = settings

        # still hardcoding this for now, hoping that cytodl api will make it unecessary
        self._cyto_dl_home_path: Path = Path(CYTO_DL_HOME_PATH)

    def get_cyto_dl_home_path(self) -> Path:
        return self._cyto_dl_home_path

    def get_user_experiments_path(self) -> Path:
        if self.settings.value(EXPERIMENTS_HOME_KEY) is None:
            return None
        else:
            return Path(self.settings.value(EXPERIMENTS_HOME_KEY))

    def set_user_experiments_path(self, path: Path):
        self.settings.setValue(EXPERIMENTS_HOME_KEY, path)

    def prompt_for_user_experiments_home(self, parent: QWidget):
        message_dialog = QMessageBox(
            parent=parent,
            text="Please select a location to store your Segmenter ML data.",
        )
        message_dialog.exec()
        path: Path = self._prompt_for_directory(parent)
        self.set_user_experiments_path(path)

    def display_change_user_experiments_home(self, parent: QWidget):
        buttonReply = QMessageBox.question(
            parent,
            "Experiments Home",
            "Experiments Home: "
            + str(self.get_user_experiments_path())
            + "\n\n Would you like to change the Experiments Home?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if buttonReply == QMessageBox.Yes:
            path: Path = self._prompt_for_directory(parent)
            self.set_user_experiments_path(path)

    def _prompt_for_directory(self, parent: QWidget) -> Path:
        directory_dialog = QFileDialog(parent=parent)
        directory_dialog.setFileMode(QFileDialog.Directory)
        file_path: str = QFileDialog.getExistingDirectory(
            parent,
            "Select a directory",
            options=QFileDialog.Option.DontUseNativeDialog
            | QFileDialog.Option.DontUseCustomDirectoryIcons
            | QFileDialog.ShowDirsOnly,
        )
        return Path(file_path)



from qtpy.QtWidgets import (
    QVBoxLayout,
    QSizePolicy,
    QLabel,
)

from allencell_ml_segmenter.sample.sample_model import SampleModel
from allencell_ml_segmenter.core.aics_widget import AicsWidget
from allencell_ml_segmenter.core.event import Event

TRAINING_NOT_RUNNING = "Training not running"
TRAINING_RUNNING = "Training running"


class SampleStateWidget(AicsWidget):
    """
    A sample widget for training a models.

    """

    def __init__(self, model: SampleModel):
        super().__init__()
        self._sample_model = model

        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._label = QLabel(TRAINING_NOT_RUNNING)
        self.layout().addWidget(self._label)

        self._sample_model.subscribe(
            Event.PROCESS_TRAINING, self, self.update_label_with_state
        )
        self._sample_model.subscribe(
            Event.PROCESS_TRAINING_SHOW_ERROR,
            self,
            self.update_label_with_error,
        )
        self._sample_model.subscribe(
            Event.PROCESS_TRAINING_CLEAR_ERROR,
            self,
            self.update_label_with_state,
        )

    def update_label_with_state(self, event):
        if self._sample_model.get_process_running():
            self._label.setText(TRAINING_RUNNING)
        else:
            self._label.setText(TRAINING_NOT_RUNNING)

    def update_label_with_error(self, event):
        self._label.setText(self._sample_model.get_error_message())

    def handle_event(self, event):
        pass



from abc import ABC, abstractmethod
from qtpy.QtCore import QRunnable


class IQRunnableManager(ABC):
    def __init__(self):
        raise RuntimeError(
            "Cannot initialize new singleton, please use .global_instance() instead"
        )

    @abstractmethod
    def run(self, runnable: QRunnable) -> None:
        pass

    @classmethod
    @abstractmethod
    def global_instance(cls):
        pass



from pathlib import Path
import pytest
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter.core.image_data_extractor import (
    FakeImageDataExtractor,
)
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel

from allencell_ml_segmenter.services.training_service import (
    TrainingService,
)
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
)
import allencell_ml_segmenter


@pytest.fixture
def experiments_model() -> ExperimentsModel:
    exp_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
    )
    experiments_model = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(), user_experiments_path=exp_path
        )
    )
    return experiments_model


@pytest.fixture
def training_model(experiments_model: ExperimentsModel) -> TrainingModel:
    model: TrainingModel = TrainingModel(MainModel(), experiments_model)
    return model


def test_service_reacts_to_image_dir_set(
    qtbot: QtBot,
    training_model: TrainingModel,
    experiments_model: ExperimentsModel,
) -> None:
    # Arrange
    service: TrainingService = TrainingService(
        training_model,
        experiments_model,
        img_data_extractor=FakeImageDataExtractor.global_instance(),
    )
    img_dir: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "multiple_csv"
    )

    # Act / Assert
    # we expect the service to listen for the signal emitted when the image directory is set
    # do some async work, then set num channels, which will emit the signal we are waiting on
    with qtbot.waitSignal(training_model.signals.num_channels_set):
        training_model.set_images_directory(img_dir)



from pathlib import Path
from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from typing import Dict, Callable, List, Optional
from allencell_ml_segmenter.core.event import Event
import allencell_ml_segmenter

test_channel_path: Path = (
    Path(allencell_ml_segmenter.__file__).parent
    / "_tests"
    / "test_files"
    / "channel_selection_json"
    / "valid_0.json"
)


class FakeExperimentsModel(IExperimentsModel):
    def __init__(
        self,
        experiments: List[str] = ["0_exp", "1_exp", "2_exp"],
        channel_selection_path: Path = test_channel_path,
    ):
        super().__init__()
        self._experiments = experiments
        self._experiment_name = None
        self._checkpoint = None
        self._events_to_subscriber_handlers: Dict[
            Dict[Subscriber, Callable]
        ] = {event: dict() for event in [e.value for e in Event]}
        self._chan_sel_path = channel_selection_path

    def get_checkpoint(self) -> str:
        return self._checkpoint

    def set_checkpoint(self, checkpoint: str):
        self._checkpoint = checkpoint

    def get_experiments(self) -> List[str]:
        return self._experiments

    def refresh_experiments(self):
        pass

    def refresh_checkpoints(self, experiment: str):
        pass

    def get_user_experiments_path(self) -> Path:
        return Path()

    def get_model_checkpoints_path(
        self, experiment_name: str, checkpoint: str
    ) -> Path:
        return Path()

    def get_metrics_csv_path(self) -> Path:
        return Path()

    def get_latest_metrics_csv_version(self) -> int:
        pass

    def get_latest_metrics_csv_path(self) -> Optional[Path]:
        return Path()

    def get_csv_path(self) -> Path:
        return Path()

    def get_cache_dir(self) -> Path:
        return Path()

    def get_channel_selection_path(self) -> Optional[Path]:
        return self._chan_sel_path



from napari.utils.notifications import show_warning
from pathlib import Path
from typing import Optional

from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.main.i_viewer import IViewer
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QLabel,
    QPushButton,
    QFrame,
    QVBoxLayout,
    QSizePolicy,
    QWidget,
    QGridLayout,
    QComboBox,
    QHBoxLayout,
    QRadioButton,
    QLineEdit,
    QCheckBox,
)
from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.view import View, MainWindow
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.training.image_selection_widget import (
    ImageSelectionWidget,
)
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ModelSize,
)
from allencell_ml_segmenter.training.patch_size_validator import (
    PatchSizeValidator,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint
from qtpy.QtGui import QIntValidator
from allencell_ml_segmenter.training.training_progress_tracker import (
    TrainingProgressTracker,
)
from allencell_ml_segmenter.core.info_dialog_box import InfoDialogBox
from allencell_ml_segmenter.utils.file_utils import FileUtils


class TrainingView(View, MainWindow):
    """
    Holds widgets pertinent to training processes - ImageSelectionWidget & ModelSelectionWidget.
    """

    def __init__(
        self,
        main_model: MainModel,
        experiments_model: IExperimentsModel,
        training_model: TrainingModel,
        viewer: IViewer,
    ):
        super().__init__()

        self._viewer: IViewer = viewer

        self._main_model: MainModel = main_model
        self._experiments_model: IExperimentsModel = experiments_model
        self._training_model: TrainingModel = training_model

        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().setAlignment(Qt.AlignTop)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)

        self._title: QLabel = QLabel("SEGMENTATION MODEL TRAINING", self)
        self._title.setObjectName("title")
        self.layout().addWidget(
            self._title, alignment=Qt.AlignHCenter | Qt.AlignTop
        )

        # initialize constituent widgets
        self.image_selection_widget: ImageSelectionWidget = (
            ImageSelectionWidget(self._training_model, self._experiments_model)
        )
        self.image_selection_widget.setObjectName("imageSelection")

        # Dummy divs allow for easy alignment
        top_container: QVBoxLayout = QVBoxLayout()
        top_dummy: QFrame = QFrame()
        bottom_dummy: QFrame = QFrame()
        top_container.addWidget(self.image_selection_widget)
        top_dummy.setLayout(top_container)
        self.layout().addWidget(top_dummy)

        # bottom half
        bottom_grid_layout = QGridLayout()

        patch_size_text_layout = QVBoxLayout()
        patch_size_text_layout.setSpacing(0)
        patch_size_label: LabelWithHint = LabelWithHint("Patch size")
        patch_size_label.set_hint(
            "Patch size to split images into during training. Should encompass the structure of interest and all dimensions should be evenly divisble by 4. If 2D, Z can be left blank."
        )
        patch_size_text_layout.addWidget(patch_size_label)
        guide_text: QLabel = QLabel("All values must be multiples of 4")
        guide_text.setObjectName("subtext")
        patch_size_text_layout.addWidget(guide_text)
        bottom_grid_layout.addLayout(patch_size_text_layout, 0, 0)
        patch_size_entry_layout: QHBoxLayout = QHBoxLayout()

        # allow only integers for the linedits below
        patch_validator: PatchSizeValidator = PatchSizeValidator()

        self.z_patch_size: QLineEdit = QLineEdit()
        self.z_patch_size.setValidator(patch_validator)
        patch_size_entry_layout.addWidget(QLabel("Z:"))
        patch_size_entry_layout.addWidget(self.z_patch_size)

        self.y_patch_size: QLineEdit = QLineEdit()
        self.y_patch_size.setValidator(patch_validator)
        patch_size_entry_layout.addWidget(QLabel("Y:"))
        patch_size_entry_layout.addWidget(self.y_patch_size)

        self.x_patch_size: QLineEdit = QLineEdit()
        self.x_patch_size.setValidator(patch_validator)
        patch_size_entry_layout.addWidget(QLabel("X:"))
        patch_size_entry_layout.addWidget(self.x_patch_size)

        bottom_grid_layout.addLayout(patch_size_entry_layout, 0, 1)

        model_size_label: LabelWithHint = LabelWithHint("Model size")
        model_size_label.set_hint(
            "Defines the complexity of the model - smaller models train faster, while large models train slower but may learn complex relationships better."
        )
        bottom_grid_layout.addWidget(model_size_label, 1, 0)

        self._model_size_combo_box: QComboBox = QComboBox()
        self._model_size_combo_box.setObjectName("modelSizeComboBox")
        self._model_size_combo_box.setCurrentIndex(-1)
        self._model_size_combo_box.setPlaceholderText("Select an option")
        self._model_size_combo_box.addItems(
            [size.name.lower() for size in ModelSize]
        )
        self._model_size_combo_box.currentTextChanged.connect(
            lambda model_size: self._training_model.set_model_size(model_size)
        )
        bottom_grid_layout.addWidget(self._model_size_combo_box, 1, 1)

        image_dimensions_label: LabelWithHint = LabelWithHint(
            "Image dimension"
        )
        image_dimensions_label.set_hint("Dimensionality of your image data")
        bottom_grid_layout.addWidget(image_dimensions_label, 2, 0)

        dimension_choice_layout: QHBoxLayout = QHBoxLayout()
        dimension_choice_layout.setSpacing(0)

        self._radio_3d: QRadioButton = QRadioButton()
        self._radio_3d.setObjectName("3DRadio")
        self._radio_3d.toggled.connect(
            lambda: self._training_model.set_spatial_dims(3)
        )
        label_3d: LabelWithHint = LabelWithHint("3D")

        self._radio_2d: QRadioButton = QRadioButton()
        self._radio_2d.toggled.connect(
            lambda: self._training_model.set_spatial_dims(2)
        )
        label_2d: LabelWithHint = LabelWithHint("2D")

        dimension_choice_layout.addWidget(self._radio_3d)
        dimension_choice_layout.addWidget(label_3d)
        dimension_choice_layout.addWidget(
            self._radio_2d, alignment=Qt.AlignLeft
        )
        dimension_choice_layout.addWidget(label_2d, alignment=Qt.AlignLeft)
        dimension_choice_layout.addStretch(10)
        dimension_choice_layout.setContentsMargins(0, 0, 0, 0)

        dimension_choice_dummy: QWidget = (
            QWidget()
        )  # stops interference with other radio buttons
        dimension_choice_dummy.setLayout(dimension_choice_layout)

        bottom_grid_layout.addWidget(dimension_choice_dummy, 2, 1)

        num_epochs_label: LabelWithHint = LabelWithHint("Number of epochs")
        num_epochs_label.set_hint(
            "Each epoch is one complete pass through the entire training dataset. More epochs yields better performance at the cost of training time."
        )
        bottom_grid_layout.addWidget(num_epochs_label, 3, 0)

        self._num_epochs_input: QLineEdit = QLineEdit()
        int_validator: QIntValidator = QIntValidator()
        int_validator.setBottom(1)
        self._num_epochs_input.setValidator(int_validator)
        self._num_epochs_input.setPlaceholderText("1000")
        self._num_epochs_input.setObjectName("trainingStepInput")
        self._num_epochs_input.textChanged.connect(
            self._num_epochs_field_handler
        )
        bottom_grid_layout.addWidget(self._num_epochs_input, 3, 1)

        max_time_layout: QHBoxLayout = QHBoxLayout()
        max_time_layout.setSpacing(0)

        self._max_time_checkbox: QCheckBox = QCheckBox()
        self._max_time_checkbox.setObjectName("timeoutCheckbox")
        self._max_time_checkbox.stateChanged.connect(
            self._max_time_checkbox_slot
        )
        max_time_layout.addWidget(self._max_time_checkbox)

        max_time_left_text: QLabel = QLabel("Time out after")
        max_time_layout.addWidget(max_time_left_text)

        self._max_time_in_minutes_input: QLineEdit = QLineEdit()
        self._max_time_in_minutes_input.setObjectName("timeoutMinuteInput")
        self._max_time_in_minutes_input.setEnabled(False)
        self._max_time_in_minutes_input.setMaximumWidth(30)
        self._max_time_in_minutes_input.setPlaceholderText("30")
        self._max_time_in_minutes_input.textChanged.connect(
            lambda text: self._training_model.set_max_time(int(text))
        )
        max_time_layout.addWidget(self._max_time_in_minutes_input)

        max_time_right_text: LabelWithHint = LabelWithHint("minutes")
        max_time_right_text.set_hint("(Optional) Maximum time to train model")
        max_time_layout.addWidget(max_time_right_text, alignment=Qt.AlignLeft)
        max_time_layout.addStretch()

        bottom_grid_layout.addLayout(max_time_layout, 4, 1)
        bottom_grid_layout.setColumnStretch(1, 8)
        bottom_grid_layout.setColumnStretch(0, 3)

        bottom_dummy.setLayout(bottom_grid_layout)
        self.layout().addWidget(bottom_dummy)

        self._train_btn: QPushButton = QPushButton("Start training")
        self._train_btn.setObjectName("trainBtn")
        self.layout().addWidget(self._train_btn)
        self._train_btn.clicked.connect(self.train_btn_handler)

        self._main_model.subscribe(
            Event.VIEW_SELECTION_TRAINING,
            self,
            lambda e: self._main_model.set_current_view(self),
        )

        # apply styling
        self.setStyleSheet(Style.get_stylesheet("training_view.qss"))

    def train_btn_handler(self) -> None:
        """
        Starts training process
        """
        if self._patch_size_ok():
            self.set_patch_size()

            progress_tracker: TrainingProgressTracker = (
                TrainingProgressTracker(
                    self._experiments_model.get_metrics_csv_path(),
                    self._experiments_model.get_cache_dir(),
                    self._training_model.get_num_epochs(),
                    self._training_model.get_total_num_images(),
                    self._experiments_model.get_latest_metrics_csv_version()
                    + 1,
                )
            )
            self.startLongTaskWithProgressBar(progress_tracker)

    # Abstract methods from View implementations #######################

    def doWork(self):
        """
        Starts training process
        """
        self._training_model.dispatch_training()

    def getTypeOfWork(self) -> str:
        """
        Returns string representation of training process
        """
        return "Training"

    def showResults(self):
        csv_path: Path = self._experiments_model.get_latest_metrics_csv_path()
        min_loss: Optional[float] = FileUtils.get_min_loss_from_csv(csv_path)
        dialog_box = InfoDialogBox(
            "Training finished -- Final loss: {:.3f}".format(min_loss)
        )
        dialog_box.exec()

    def _num_epochs_field_handler(self, num_epochs: str) -> None:
        self._training_model.set_num_epochs(int(num_epochs))

    def _max_time_checkbox_slot(self, checked: Qt.CheckState) -> None:
        """
        Triggered when the user selects the "time out after" _timeout_checkbox.
        Enables/disables interaction with the neighboring hour input based on checkstate.
        """
        if checked == Qt.CheckState.Checked:
            self._max_time_in_minutes_input.setEnabled(True)
            self._training_model.set_use_max_time(True)
        else:
            self._max_time_in_minutes_input.setEnabled(False)
            self._training_model.set_use_max_time(False)

    def _patch_size_ok(self) -> bool:
        """
        Validates that user has defined all needed patch sizes in the UI
        Returns True if all needed patch sizes were provided, False if not (and will show napari error message)
        """
        missing_patches: list[str] = []

        if (
            not self.z_patch_size.text()
            and self._training_model.get_spatial_dims() == 3
        ):
            missing_patches.append("Z")

        if not self.y_patch_size.text():
            missing_patches.append("Y")

        if not self.x_patch_size.text():
            missing_patches.append("X")

        if len(missing_patches) > 0:
            show_warning(
                f"Please define {missing_patches} patches before continuing."
            )
            return False

        return True

    def set_patch_size(self) -> None:
        self._training_model.set_patch_size(
            [
                int(self.z_patch_size.text()),
                int(self.y_patch_size.text()),
                int(self.x_patch_size.text()),
            ]
        )

    def focus_changed(self) -> None:
        self.image_selection_widget.set_inputs_csv()
        self._viewer.clear_layers()



import pytest

from allencell_ml_segmenter._style import Style


class TestStyle:
    @pytest.mark.parametrize("name", [None, "not_a_qt_stylesheet.txt"])
    def test_load_stylesheet_bad_name(self, name: str) -> None:
        """
        Tests that an error is raised when trying to load a stylesheet with an invalid name.
        """
        with pytest.raises(ValueError):
            Style.get_stylesheet(name)

    def test_load_stylesheet_bad_path(self) -> None:
        """
        Tests that an error is raised when trying to load a stylesheet that doesn't exist.
        """
        with pytest.raises(IOError):
            Style.get_stylesheet("not_found.qss")

    def test_load_stylesheet(self) -> None:
        """
        Tests that an existing stylesheet is properly loaded.
        """
        contents = Style.get_stylesheet("prediction_view.qss")
        assert contents is not None
        assert len(contents) > 0



from abc import ABC, abstractmethod
from pathlib import Path


class IExtractorFactory(ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def create(self, img_path: Path):
        pass



import pytest
from pathlib import Path
import allencell_ml_segmenter
from allencell_ml_segmenter.core.channel_extraction import *


def test_extract_channels_from_image():
    img: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "images"
        / "test_3_channels.tiff"
    )
    assert extract_channels_from_image(img) == 3


def test_get_img_path_from_csv():
    csv_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "csv"
        / "test_csv.csv"
    )
    img_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "images"
        / "test_3_channels.tiff"
    )
    assert get_img_path_from_csv(csv_path).samefile(img_path)



from allencell_ml_segmenter._tests.fakes.fake_qsettings import FakeQSettings

from allencell_ml_segmenter.config.user_settings import UserSettings


def test_set_get_user_experiments_path():
    # Arrange
    settings = FakeQSettings()
    userSettings = UserSettings(settings=settings)

    # Act
    userSettings.set_user_experiments_path("foo")

    # Assert
    userSettings.get_user_experiments_path() == "fooo"
